try:
    _HAVE_MATPLOTLIB = True
    import matplotlib.pyplot as plt
except ImportError:
    _HAVE_MATPLOTLIB = False

import numpy as np
from copy import deepcopy
import os
from scipy.spatial import KDTree

try:
    import pyvista as pv

    _HAVE_PYVISTA = True
except ImportError:
    _HAVE_PYVISTA = False

from zoomy_core.mesh.fvm_mesh import Mesh
from zoomy_core.misc.io import _save_fields_to_hdf5 as save_fields_to_hdf5
import zoomy_core.misc.io as io


main_dir = os.getenv("ZOOMY_DIR")


def load_file(filename):
    if not _HAVE_PYVISTA:
        raise ImportError("pyvista is required for load_file function.")
    reader = pv.get_reader(filename)
    # get(0) gets us the internal (not boundary) data. The boundary data is non-existant anyways in our case
    vtkfile = reader.read().get(0)
    return vtkfile


def get_fields(vtkfile, fieldnames):
    fields = []
    for fieldname in fieldnames:
        # point field)s
        # field = vtkfile.point_data[fieldname]
        field = vtkfile.cell_data[fieldname]
        fields.append(np.array(field))
    return fields
    # return np.array(fields)


def get_coordinates(vtkfile):
    # point coordinates
    # return np.array(vtkfile.points)
    # cell centers
    return vtkfile.cell_centers().points


def get_time(vtkfile):
    return vtkfile.field_data["TimeValue"][0]


def sort_data(coordinates, fields):
    # Sort by z, y, x
    order = np.lexsort((coordinates[:, 0], coordinates[:, 1], coordinates[:, 2]))
    return order


def apply_order(coordinates, fields, order):
    coordinates_copy = np.array(coordinates)
    fields_copy = deepcopy(fields)
    for d in range(coordinates.shape[1]):
        coordinates[:, d] = coordinates_copy[order, d]
    for field, field_copy in zip(fields, fields_copy):
        field[:] = field_copy[order]
    return coordinates, fields


def get_number_of_layers_and_elements_in_plane(coordinates):
    layers = len(np.unique(coordinates[:, 2]))
    n_elements_plane = int(coordinates[:, 0].size / layers)
    return layers, n_elements_plane


def get_layer(data, layer, n_elements_plane):
    return data[layer * n_elements_plane : (layer + 1) * n_elements_plane]


def compute_height(
    alpha, n_layers, n_elements_per_layer, total_height=1.0, threshold=0.5
):
    height = np.zeros(n_elements_per_layer)
    active = np.ones(n_elements_per_layer, dtype=bool)
    dh = total_height / n_layers
    for i in range(n_layers):
        alpha_layer = get_layer(alpha, i, n_elements_per_layer)
        height += np.where(np.logical_and((alpha_layer >= threshold), active), dh, 0)
    return height


def extract_faces(mesh):
    faces = []
    i, offset = 0, 0
    cc = mesh.cells  # fetch up front
    while i < mesh.n_cells:
        nn = cc[offset]
        faces.append(cc[offset + 1 : offset + 1 + nn])
        offset += nn + 1
        i += 1
    return np.array(faces)


def sort_faces(faces, order):
    for face in faces:
        for i, point in enumerate(face):
            face[i] = order[point]
    faces = faces[order]
    return faces


def add_noslip_layer(U, n_elements_per_layer):
    Unew = np.zeros((U.shape[0] + n_elements_per_layer, U.shape[1]))
    for d in range(U.shape[1]):
        Unew[n_elements_per_layer:, d] = U[:, d]
    return Unew


def extract_velocity_column_at_coordinate(U, n_elements_per_layer, n_layer, coordinate):
    u = np.array(
        [
            get_layer(U[:, 0], i, n_elements_per_layer)[coordinate]
            for i in range(n_layer + 1)
        ]
    )
    v = np.array(
        [
            get_layer(U[:, 1], i, n_elements_per_layer)[coordinate]
            for i in range(n_layer + 1)
        ]
    )
    w = np.array(
        [
            get_layer(U[:, 2], i, n_elements_per_layer)[coordinate]
            for i in range(n_layer + 1)
        ]
    )
    return u, v, w


def shift_integration_interval(xi):
    return (xi + 1) / 2


def plot_basis(basis_generator):
    fig, ax = plt.subplots()
    basis = [basis_generator(n) for n in range(0, 8)]
    X = np.linspace(0, 1, 100)
    for i in range(8):
        ax.plot(basis[i](X), X)
    return fig, ax


def moment_projection(field, n_layers, basis, integration_order=None):
    if integration_order is None:
        integration_order = max(len(basis), n_layers)
    xi, wi = np.polynomial.legendre.leggauss(integration_order)
    xi = shift_integration_interval(xi)
    dz = 1 / (n_layers)
    xp = np.arange(dz / 2, 1 - dz / 2, dz)
    xp = np.insert(xp, 0, 0)
    fp = field
    field_xi = np.interp(xi, xp, fp)
    basis_xi = [basis[i](xi) for i in range(len(basis))]
    projections = np.zeros(len(basis))
    for i in range(len(basis)):
        projections[i] = np.sum(field_xi * basis_xi[i] * wi)
    return projections


def convert_openfoam_to_moments_single(filename, n_levels):
    vtkfile = load_file(filename)
    coordinates = get_coordinates(vtkfile)
    n_layer, n_elements_per_layer = get_number_of_layers_and_elements_in_plane(
        coordinates
    )
    fields = get_fields(vtkfile, ["alpha.water", "U"])
    sort_order = sort_data(coordinates, fields)
    coordinates, fields = apply_order(coordinates, fields, sort_order)
    fields[1] = add_noslip_layer(fields[1], n_elements_per_layer)
    time = get_time(vtkfile)
    Q, basis = compute_shallow_moment_projection(fields, coordinates, n_levels)
    return coordinates, Q, time, basis


def convert_openfoam_to_moments(
    filepath, n_levels, filepath_mesh_for_order, meshtype_order="triganle"
):
    filepath_vtk = os.path.join(filepath, "VTK")
    filename = "fields_openfoam.hdf5"
    sort_order = None

    # file order
    file_number_list = []
    file_start = ""
    for i_file, file in enumerate(os.listdir(filepath_vtk)):
        if file.endswith(".vtm"):
            file_start = file.split("_")[0]
            file_start = file.rsplit("_", 1)[0]
            index = int(file.rsplit("_", 1)[1].split(".")[0])
            # index = int(re.findall(r'\d+', file)[-1])
            file_number_list.append(index)

    file_number_list.sort()
    print(file_number_list)

    # first iteration
    file0 = (
        os.path.join(filepath_vtk, file_start) + "_" + str(file_number_list[0]) + ".vtm"
    )
    coordinates, Q, time, basis = convert_openfoam_to_moments_single(file0, n_levels)
    mesh_order = Mesh.load_gmsh(filepath_mesh_for_order, meshtype_order)
    Q, sort_order = sort_fields_by_mesh(mesh_order, coordinates, Q)
    if os.path.exists(os.path.join(filepath, filename)):
        os.remove(os.path.join(filepath, filename))
    save_fields_to_hdf5(filepath, 0, time, Q.T, Qaux=None, filename=filename)

    # loop iterations
    iter = 1
    total = len(file_number_list)
    print(f"Conversion {iter}/{total} completed.")
    iter += 1
    # loop iterations
    for i, i_file in enumerate(file_number_list[1:]):
        file = os.path.join(filepath_vtk, file_start) + "_" + str(i_file) + ".vtm"
        coordinates, Q, time, basis = convert_openfoam_to_moments_single(file, n_levels)
        Q = apply_order_to_fields(sort_order, Q)
        save_fields_to_hdf5(filepath, i + 1, time, Q.T, Qaux=None, filename=filename)
        print(f"Conversion {iter}/{total} completed.")
        iter += 1


def sort_fields_by_mesh(mesh, coordinates, Q):
    # coordinates are still 3d, while Q is 2d and the mesh is 2d
    n_layers, n_elements_per_layer = get_number_of_layers_and_elements_in_plane(
        coordinates
    )
    coords_2d = coordinates[:n_elements_per_layer, :2]
    coords_ordered = mesh.element_center
    # Construct a KDTree from coords_2d
    tree = KDTree(coords_2d)

    # Find the indices of the nearest points in coords_2d to the points in coords_ordered
    _, indices = tree.query(coords_ordered)

    # Use these indices to sort Q and coords_2d
    Q_sorted = Q[:, indices]
    return Q_sorted, indices


def apply_order_to_fields(order, Q):
    return Q[:, order]


def plot_contour(coordinates, field):
    if not _HAVE_MATPLOTLIB:
        raise ImportError("matplotlib is required for plot_contour function.")
    fig, ax = plt.subplots()
    n_layer, n_elements_per_layer = get_number_of_layers_and_elements_in_plane(
        coordinates
    )
    X = get_layer(coordinates[:, 0], 0, n_elements_per_layer)
    Y = get_layer(coordinates[:, 1], 0, n_elements_per_layer)
    # Z = get_layer(fields[1][:, 0], 1, n_elements_per_layer)
    # Z = get_layer(fields[0], 12, n_elements_per_layer)
    # Z = compute_height(fields[0], n_layer, n_elements_per_layer)
    Z = field
    colorbar = ax.tricontourf(X, Y, Z)
    ax.set_aspect("equal")

    circle = plt.Circle((0.5, 1), radius=0.2, fc="silver", zorder=10, edgecolor="k")
    plt.gca().add_patch(circle)
    fig.colorbar(colorbar)
    return fig, ax


def basis_legendre(i):
    def f(x):
        basis = np.polynomial.legendre.Legendre.basis(i, domain=[0, 1], window=[-1, 1])
        return basis(x) * basis(0)

    return f


def compute_shallow_moment_projection(
    fields, coordinates, n_levels, basis_generator=basis_legendre
):
    n_layers, n_elements_per_layer = get_number_of_layers_and_elements_in_plane(
        coordinates
    )
    height = compute_height(fields[0], n_layers, n_elements_per_layer)
    basis = [basis_generator(n) for n in range(0, n_levels + 1)]
    alphas = np.zeros((n_elements_per_layer, n_levels + 1))
    betas = np.zeros((n_elements_per_layer, n_levels + 1))
    for i in range(n_elements_per_layer):
        u, v, w = extract_velocity_column_at_coordinate(
            fields[1], n_elements_per_layer, n_layers, i
        )
        alphas[i] = moment_projection(u, n_layers + 1, basis)
        betas[i] = moment_projection(v, n_layers + 1, basis)
    return np.concatenate(
        (height.reshape((n_elements_per_layer, 1)), alphas, betas), axis=1
    ).T, basis


def plot_data_vs_moments(fields, coordinates, Q, basis, coordinate_index):
    if _HAVE_MATPLOTLIB is False:
        raise ImportError("matplotlib is required for plot_data_vs_moments function.")
    n_layers, n_elements_per_layer = get_number_of_layers_and_elements_in_plane(
        coordinates
    )
    n_levels = int((Q.shape[0] - 1) / 2) - 1
    X = np.linspace(0, 1, 100)
    U = np.array(
        [basis[i](X) * Q[1 + i, coordinate_index] for i in range(len(basis))]
    ).sum(axis=0)
    V = np.array(
        [
            basis[i](X) * Q[1 + n_levels + 1 + i, coordinate_index]
            for i in range(len(basis))
        ]
    ).sum(axis=0)
    dz = 1 / (n_layers)
    x = np.linspace(dz / 2, 1 - dz / 2, n_layers)
    x = np.insert(x, 0, 0)
    u, v, w = extract_velocity_column_at_coordinate(
        fields[1], n_elements_per_layer, n_layers, coordinate_index
    )
    fig, ax = plt.subplots(2)
    ax[0].plot(U, X)
    ax[0].plot(u, x)
    ax[1].plot(V, X)
    ax[1].plot(v, x)
    return fig, ax


def load_openfoam_file(filepath):
    vtkfile = load_file(filepath)
    coordinates = get_coordinates(vtkfile)
    n_layer, n_elements_per_layer = get_number_of_layers_and_elements_in_plane(
        coordinates
    )
    fields = get_fields(vtkfile, ["alpha.water", "U"])
    sort_order = sort_data(coordinates, fields)
    coordinates, fields = apply_order(coordinates, fields, sort_order)
    fields[1] = add_noslip_layer(fields[1], n_elements_per_layer)
    time = get_time(vtkfile)
    return coordinates, fields, time


def test_load():
    filepath = os.path.join(
        os.path.join(main_dir, "openfoam_data/channelflow_coarse"),
        "channelflow_coarse_0.vtm",
    )
    coordinates, fields, time = load_openfoam_file(filepath)


def test_moment_projection():
    filepath = os.path.join(
        os.path.join(main_dir, "openfoam_data/channelflow_coarse"),
        "channelflow_coarse_0.vtm",
    )
    coordinates, fields, time = load_openfoam_file(filepath)
    Q, basis = compute_shallow_moment_projection(fields, coordinates, 3)


def test_convert_openfoam_single():
    filepath = os.path.join(
        os.path.join(main_dir, "openfoam_data/channelflow_coarse"),
        "channelflow_coarse_0.vtm",
    )
    X, Q, t, basis = convert_openfoam_to_moments_single(filepath, 3)


def test_plots():
    if _HAVE_MATPLOTLIB is False:
        raise ImportError("matplotlib is required for test_plots function.")    
    filepath = os.path.join(
        os.path.join(main_dir, "openfoam_data/channelflow_coarse"),
        "channelflow_coarse_0.vtm",
    )
    coordinates, fields, time = load_openfoam_file(filepath)
    X, Q, t, basis = convert_openfoam_to_moments_single(filepath, 3)
    fig, ax = plot_data_vs_moments(fields, coordinates, Q, basis, 0)
    plt.show()
    fig, ax = plot_basis(basis_legendre)
    plt.show()
    fig, ax = plot_contour(coordinates, Q[0])
    plt.show()


def test_sort():
    filepath = os.path.join(
        os.path.join(main_dir, "openfoam_data/channelflow_coarse"),
        "channelflow_coarse_0.vtm",
    )
    filepath_gmsh = os.path.join(
        os.path.join(main_dir, "meshes/channel_openfoam/mesh_coarse_2d.msh")
    )

    mesh_gmsh = Mesh.load_gmsh(filepath_gmsh, "triangle")
    mesh_gmsh.write_to_hdf5(
        os.path.join(os.path.join(main_dir, "openfoam_data/channelflow_coarse"))
    )

    filepath_hdf_mesh = os.path.join(
        os.path.join(main_dir, "openfoam_data/channelflow_coarse/mesh.hdf5")
    )
    mesh = Mesh.from_hdf5(filepath_hdf_mesh)

    coordinates, fields, time = load_openfoam_file(filepath)
    Q, basis = compute_shallow_moment_projection(fields, coordinates, 3)

    # mesh_comparison = Mesh.load_gmsh(os.path.join(main_dir, 'meshes/channel_openfoam/mesh_coarse_3d.msh'), 'tetra')

    sort_fields_by_mesh(mesh, coordinates, Q)


def test_convert_openfoam_to_moments(level=0):
    # filepath = os.path.join(main_dir, 'openfoam_data/channelflow_coarse')
    foam_sim = os.getenv("FOAM_SIM")
    filepath = os.path.join(foam_sim, "multiphase/interFoam/RAS/channelflow_mid")
    filepath_target_mesh = os.path.join(
        os.path.join(main_dir, "meshes/simple_openfoam/mesh_2d_mid.msh")
    )
    convert_openfoam_to_moments(
        filepath, level, filepath_target_mesh, meshtype_order="triangle"
    )


if __name__ == "__main__":
    # test_load()
    # test_moment_projection()
    # test_convert_openfoam_single()
    # test_plots()
    # test_sort()
    test_convert_openfoam_to_moments(level=1)
    filepath = os.path.join(main_dir, "openfoam_data/channelflow_mid")
    filepath_mesh = os.path.join(main_dir, "meshes/simple_openfoam/mesh_2d_mid.msh")
    io.generate_vtk(
        filepath,
        filepath_gmsh=filepath_mesh,
        gmsh_mesh_type="triangle",
        filename_fields="fields_openfoam.hdf5",
        filename_out="fields_openfoam_vtk",
        skip_aux=True,
    )



from typing import Callable, Union

import numpy as np
import sympy
from attrs import define, field
from sympy import init_printing, powsimp, zeros


from zoomy_core.model.boundary_conditions import BoundaryConditions
from zoomy_core.model.initial_conditions import Constant, InitialConditions
from zoomy_core.misc.custom_types import FArray
from zoomy_core.misc.misc import Zstruct, ZArray
from zoomy_core.model.basefunction import Function

init_printing()


def default_simplify(expr):
    return powsimp(expr, combine="all", force=False, deep=True)


@define(frozen=True, slots=True, kw_only=True)
class Model:
    """
    Generic (virtual) model implementation.
    """

    boundary_conditions: BoundaryConditions

    name: str = "Model"
    dimension: int = 1

    initial_conditions: InitialConditions = field(factory=Constant)
    aux_initial_conditions: InitialConditions = field(factory=Constant)

    parameters: Zstruct = field(factory=lambda: Zstruct())

    time: sympy.Symbol = field(
        init=False, factory=lambda: sympy.symbols("t", real=True)
    )
    distance: sympy.Symbol = field(
        init=False, factory=lambda: sympy.symbols("dX", real=True)
    )
    position: Zstruct = field(
        init=False, factory=lambda: register_sympy_attribute(3, "X")
    )
    number_of_points_3d: int = field(factory=lambda: 10)

    _simplify: Callable = field(factory=lambda: default_simplify)

    # Derived fields initialized in __attrs_post_init__
    _default_parameters: dict = field(init=False, factory=dict)
    n_variables: int = field(init=False)
    n_aux_variables: int = field(init=False)
    n_parameters: int = field(init=False)
    variables: Zstruct = field(init=False, default=1)

    positive_variables: Union[dict, list, None] = field(default=None)
    aux_variables: Zstruct = field(default=0)
    parameter_values: FArray = field(init=False)
    normal: ZArray = field(init=False)

    z_3d: Zstruct = field(init=False, default=number_of_points_3d)
    u_3d: Zstruct = field(init=False, default=number_of_points_3d)
    p_3d: Zstruct = field(init=False, default=number_of_points_3d)
    alpha_3d: Zstruct = field(init=False, default=number_of_points_3d)

    _flux: Function = field(init=False)
    _dflux: Function = field(init=False)
    _nonconservative_matrix: Function = field(init=False)
    _quasilinear_matrix: Function = field(init=False)
    _eigenvalues: Function = field(init=False)
    _left_eigenvectors: Function = field(init=False)
    _right_eigenvectors: Function = field(init=False)
    _source: Function = field(init=False)
    _source_jacobian_wrt_variables: Function = field(init=False)
    _source_jacobian_wrt_aux_variables: Function = field(init=False)
    _residual: Function = field(init=False)
    _project_2d_to_3d: Function = field(init=False)
    _project_3d_to_2d: Function = field(init=False)
    _boundary_conditions: Function = field(init=False)

    def __attrs_post_init__(self):
        updated_default_parameters = {
            **self._default_parameters,
            **self.parameters.as_dict(),
        }

        # Use object.__setattr__ because class is frozen
        object.__setattr__(
            self,
            "variables",
            register_sympy_attribute(self.variables, "q", self.positive_variables),
        )
        object.__setattr__(
            self, "aux_variables", register_sympy_attribute(self.aux_variables, "qaux")
        )

        object.__setattr__(
            self,
            "parameters",
            register_sympy_attribute(updated_default_parameters, "p"),
        )
        object.__setattr__(
            self,
            "parameter_values",
            register_parameter_values(updated_default_parameters),
        )
        object.__setattr__(
            self,
            "normal",
            register_sympy_attribute(
                ["n" + str(i) for i in range(self.dimension)], "n"
            ),
        )

        object.__setattr__(self, "n_variables", self.variables.length())
        object.__setattr__(self, "n_aux_variables", self.aux_variables.length())
        object.__setattr__(self, "n_parameters", self.parameters.length())

        object.__setattr__(
            self, "z_3d", register_sympy_attribute(self.number_of_points_3d, "z")
        )
        object.__setattr__(
            self, "u_3d", register_sympy_attribute(self.number_of_points_3d, "u")
        )
        object.__setattr__(
            self, "p_3d", register_sympy_attribute(self.number_of_points_3d, "p")
        )
        object.__setattr__(
            self,
            "alpha_3d",
            register_sympy_attribute(self.number_of_points_3d, "alpha"),
        )

        object.__setattr__(
            self,
            "_flux",
            Function(
                name="flux",
                definition=self.flux(),
                args=Zstruct(
                    variables=self.variables,
                    aux_variables=self.aux_variables,
                    parameters=self.parameters,
                ),
            ),
        )
        object.__setattr__(
            self,
            "_dflux",
            Function(
                name="dflux",
                definition=self.dflux(),
                args=Zstruct(
                    variables=self.variables,
                    aux_variables=self.aux_variables,
                    parameters=self.parameters,
                ),
            ),
        )

        object.__setattr__(
            self,
            "_nonconservative_matrix",
            Function(
                name="nonconservative_matrix",
                definition=self.nonconservative_matrix(),
                args=Zstruct(
                    variables=self.variables,
                    aux_variables=self.aux_variables,
                    parameters=self.parameters,
                ),
            ),
        )
        object.__setattr__(
            self,
            "_quasilinear_matrix",
            Function(
                name="quasilinear_matrix",
                definition=self.quasilinear_matrix(),
                args=Zstruct(
                    variables=self.variables,
                    aux_variables=self.aux_variables,
                    parameters=self.parameters,
                ),
            ),
        )
        object.__setattr__(
            self,
            "_eigenvalues",
            Function(
                name="eigenvalues",
                definition=self.eigenvalues(),
                args=Zstruct(
                    variables=self.variables,
                    aux_variables=self.aux_variables,
                    parameters=self.parameters,
                    normal=self.normal,
                ),
            ),
        )
        object.__setattr__(
            self,
            "_left_eigenvectors",
            Function(
                name="left_eigenvectors",
                definition=self.left_eigenvectors(),
                args=Zstruct(
                    variables=self.variables,
                    aux_variables=self.aux_variables,
                    parameters=self.parameters,
                    normal=self.normal,
                ),
            ),
        )
        object.__setattr__(
            self,
            "_right_eigenvectors",
            Function(
                name="right_eigenvectors",
                definition=self.right_eigenvectors(),
                args=Zstruct(
                    variables=self.variables,
                    aux_variables=self.aux_variables,
                    parameters=self.parameters,
                    normal=self.normal,
                ),
            ),
        )
        object.__setattr__(
            self,
            "_source",
            Function(
                name="source",
                definition=self.source(),
                args=Zstruct(
                    variables=self.variables,
                    aux_variables=self.aux_variables,
                    parameters=self.parameters,
                ),
            ),
        )
        object.__setattr__(
            self,
            "_source_jacobian_wrt_variables",
            Function(
                name="source_jacobian_wrt_variables",
                definition=self.source_jacobian_wrt_variables(),
                args=Zstruct(
                    variables=self.variables,
                    aux_variables=self.aux_variables,
                    parameters=self.parameters,
                ),
            ),
        )
        object.__setattr__(
            self,
            "_source_jacobian_wrt_aux_variables",
            Function(
                name="source_jacobian_wrt_aux_variables",
                definition=self.source_jacobian_wrt_aux_variables(),
                args=Zstruct(
                    variables=self.variables,
                    aux_variables=self.aux_variables,
                    parameters=self.parameters,
                ),
            ),
        )

        object.__setattr__(
            self,
            "_residual",
            Function(
                name="residual",
                definition=self.residual(),
                args=Zstruct(
                    time=self.time,
                    position=self.position,
                    distance=self.distance,
                    variables=self.variables,
                    aux_variables=self.aux_variables,
                    parameters=self.parameters,
                ),
            ),
        )
        object.__setattr__(
            self,
            "_project_2d_to_3d",
            Function(
                name="project_2d_to_3d",
                definition=self.project_2d_to_3d(),
                args=Zstruct(
                    Z=self.position[2],
                    variables=self.variables,
                    aux_variables=self.aux_variables,
                    parameters=self.parameters,
                ),
            ),
        )
        object.__setattr__(
            self,
            "_project_3d_to_2d",
            Function(
                name="project_3d_to_2d",
                definition=self.project_3d_to_2d(),
                args=Zstruct(
                    variables=self.variables,
                    aux_variables=self.aux_variables,
                    parameters=self.parameters,
                ),
            ),
        )
        object.__setattr__(
            self,
            "_boundary_conditions",
            self.boundary_conditions.get_boundary_condition_function(
                self.time,
                self.position,
                self.distance,
                self.variables,
                self.aux_variables,
                self.parameters,
                self.normal,
            ),
        )

    def flux(self):
        return ZArray.zeros(
            self.n_variables,
            self.dimension,
        )

    def dflux(self):
        return ZArray.zeros(
            self.n_variables,
            self.dimension,
        )

    def nonconservative_matrix(self):
        return ZArray.zeros(
            self.n_variables,
            self.n_variables,
            self.dimension,
        )

    def source(self):
        return ZArray.zeros(self.n_variables)
        return zeros(self.n_variables, 1)

    def quasilinear_matrix(self):
        """generated automatically unless explicitly provided"""
        # NC = [ZArray.zeros(self.n_variables, self.n_variables) for _ in range(self.dimension)]
        # NC = ZArray.zeros(
        #     self.n_variables, self.n_variables, self.dimension
        # )
        JacF = ZArray(sympy.derive_by_array(self.flux(), self.variables.get_list()))
        for d in range(self.dimension):
            JacF_d = JacF[:, :, d]
            JacF_d = ZArray(JacF_d.tomatrix().T)
            JacF[:, :, d] = JacF_d
        return self._simplify(
            JacF + self.nonconservative_matrix()
        )

    def source_jacobian_wrt_variables(self):
        """generated automatically unless explicitly provided"""
        return self._simplify(
            sympy.derive_by_array(self.source(), self.variables.get_list())
        )

    def source_jacobian_wrt_aux_variables(self):
        """generated automatically unless explicitly provided"""
        return self._simplify(
            sympy.derive_by_array(self.source(), self.aux_variables.get_list())
        )

    def residual(self):
        return ZArray.zeros(self.n_variables)

    def project_2d_to_3d(self):
        return ZArray.zeros(6)

    def project_3d_to_2d(self):
        return ZArray.zeros(self.n_variables)

    def eigenvalues(self):
        A = self.normal[0] * self.quasilinear_matrix()[:, :, 0]
        for d in range(1, self.dimension):
            A += self.normal[d] * self.quasilinear_matrix()[:, :, d]
        return self._simplify(eigenvalue_dict_to_matrix(sympy.Matrix(A).eigenvals()))

    def left_eigenvectors(self):
        return ZArray.zeros(self.n_variables, self.n_variables)

    def right_eigenvectors(self):
        return ZArray.zeros(self.n_variables, self.n_variables)


def substitute_precomputed_denominator(self, expr, sym, sym_inv):
    """Recursively replace denominators involving `sym` with precomputed `sym_inv`.
    Works for scalar, Matrix, or Array of any rank.
    """
    # --- Case 1: Matrix (2D) ---
    if isinstance(expr, sympy.MatrixBase):
        return expr.applyfunc(
            lambda e: self.substitute_precomputed_denominator(e, sym, sym_inv)
        )

    # --- Case 2: Array (any rank) ---
    if isinstance(expr, sympy.Array):
        # Recursively apply substitution elementwise
        new_data = [
            self.substitute_precomputed_denominator(e, sym, sym_inv) for e in expr
        ]
        return sympy.ZArray(new_data).reshape(*expr.shape)

    # --- Case 3: Scalar or general SymPy expression ---
    num, den = sympy.fraction(expr)

    if den.has(sym):
        # split denominator into sym-dependent and independent parts
        den_sym, den_rest = den.as_independent(sym, as_Add=False)
        # swap naming (as_independent returns independent first)
        den_rest, den_sym = den_sym, den_rest

        # replace sym by sym_inv in the sym-dependent part
        den_sym_repl = den_sym.xreplace({sym: sym_inv})

        return (
            self.substitute_precomputed_denominator(num, sym, sym_inv)
            * den_sym_repl
            / den_rest
        )

    # recurse through function arguments
    elif hasattr(expr, "args") and expr.args:
        return expr.func(
            *[
                self.substitute_precomputed_denominator(arg, sym, sym_inv)
                for arg in expr.args
            ]
        )

    # base case: atomic expression
    else:
        return expr


def transform_positive_variable_intput_to_list(argument, positive, n_variables):
    out = [False for _ in range(n_variables)]
    if positive is None:
        return out
    if type(positive) == type({}):
        assert type(argument) == type(positive)
        for i, a in enumerate(argument.keys()):
            if a in positive.keys():
                out[i] = positive[a]
    if type(positive) == list:
        for i in positive:
            out[i] = True
    return out


def register_sympy_attribute(argument, string_identifier="q_", positives=None):
    if type(argument) == int:
        positive = transform_positive_variable_intput_to_list(
            argument, positives, argument
        )
        attributes = {
            string_identifier + str(i): sympy.symbols(
                string_identifier + str(i), real=True, positive=positive[i]
            )
            for i in range(argument)
        }
    elif type(argument) == type({}):
        positive = transform_positive_variable_intput_to_list(
            argument, positives, len(argument)
        )
        attributes = {
            name: sympy.symbols(str(name), real=True, positive=pos)
            for name, pos in zip(argument.keys(), positive)
        }
    elif type(argument) == list:
        positive = transform_positive_variable_intput_to_list(
            argument, positives, len(argument)
        )
        attributes = {
            name: sympy.symbols(str(name), real=True, positive=pos)
            for name, pos in zip(argument, positive)
        }
    else:
        assert False
    return Zstruct(**attributes)


def register_parameter_values(parameters):
    if type(parameters) == int:
        default_values = np.zeros(parameters, dtype=float)
    elif type(parameters) == type({}):
        default_values = np.array([value for value in parameters.values()])
    else:
        assert False
    return default_values


def eigenvalue_dict_to_matrix(eigenvalues, simplify=default_simplify):
    evs = []
    for ev, mult in eigenvalues.items():
        for i in range(mult):
            evs.append(simplify(ev))
    return ZArray(evs)



from importlib.metadata import version, PackageNotFoundError

try:
    __version__ = version("zoomy-core")
except PackageNotFoundError:
    # Package not installed, e.g. running from source
    __version__ = "0.0.0"



import numpy as np
import numpy.polynomial.legendre as L
import numpy.polynomial.chebyshev as C
from scipy.optimize import least_squares as lsq
import sympy
from sympy import Matrix, Piecewise
from sympy.abc import x

from sympy import integrate, diff
from sympy import legendre
from sympy import lambdify

from attrs import define, field
import attr
from typing import Union, Dict, List

from sympy.functions.elementary.exponential import LambertW


from zoomy_core.model.basemodel import (
    register_sympy_attribute,
    eigenvalue_dict_to_matrix,
)
from zoomy_core.model.basemodel import Model
from zoomy_core.model.models.basismatrices import Basismatrices
from zoomy_core.model.models.basisfunctions import Legendre_shifted, Basisfunction

from sympy.integrals.quadrature import gauss_legendre


    

@define(frozen=True, slots=True, kw_only=True)
class SMET(Model):
    dimension: int = 2
    level: int
    variables: Union[list, int] = field(init=False)
    positive_variables: Union[List[int], Dict[str, int], None] = attr.ib(default=attr.Factory(lambda: [1]))    
    aux_variables: Union[list, int] = field(default=0)
    basisfunctions: Union[Basisfunction, type[Basisfunction]] = field(default=Legendre_shifted)
    basismatrices: Basismatrices = field(init=False)
    order_numerical_integration: int = 4

    _default_parameters: dict = field(
        init=False,
        factory=lambda: {"g": 9.81, "ex": 0.0, "ey": 0.0, "ez": 1.0, "eps_low_water": 1e-6, "rho": 1000., 'nu': 1e-6, 'kappa': 0.41, 'B': 5.2, 'Cs': 0.17, 'yp': 0.1, 'CsW': 0.5, 'Ks': 0.001},
    )

    def __attrs_post_init__(self):
        object.__setattr__(self, "variables", ((self.level+1)*self.dimension)+2)
        # [dQ_dx, dQ_dy, deltaX, user-defined]
        object.__setattr__(self, "aux_variables", 2*((self.level+1)*self.dimension+2)+1+self.aux_variables)
        super().__attrs_post_init__()
        aux_variables = self.aux_variables
        aux_var_list = aux_variables.keys()
        object.__setattr__(self, "aux_variables", register_sympy_attribute(aux_var_list, "qaux_"))

        # Recompute basis matrices
        object.__setattr__(self, "basisfunctions", self.basisfunctions(level=self.level))
        basismatrices = Basismatrices(self.basisfunctions)
        basismatrices.compute_matrices(self.level)
        object.__setattr__(self, "basismatrices", basismatrices)

    def get_primitives(self):
        offset = self.level + 1
        b = self.variables[0]
        h = self.variables[1]
        hinv = 1/h
        ha = self.variables[2 : 2 + self.level + 1]
        alpha = [ha[i] * hinv for i in range(offset)]
        if self.dimension == 1:
            hb = [0 for i in range(self.level+1)]
        else:
            hb = self.variables[2 + offset : 2 + offset + self.level + 1]
        beta = [hb[i] * hinv for i in range(offset)]
        return [b, h, alpha, beta, hinv]

    def get_gradient(self):
        offset = self.level + 1
        z = sympy.Symbol('z')
        phi = [self.basisfunctions.eval(k, z) for k in range(self.level+1)]
        
        
        if self.dimension == 1:
            grad_b = [self.aux_variables[0]]
            grad_h = [self.aux_variables[1]]
            dalpha_dx = self.aux_variables[2:2+offset]
            dU_dx = sum([dalpha_dx[k] * phi[k] for k in range(self.level+1)])
            grad_alpha = [dalpha_dx]
            grad_beta = [0 for i in range(offset)]
            grad_U = [[dU_dx, 0], [0, 0]]
        elif self.dimension == 2:
            gradient_offset = self.n_variables
            grad_b = [self.aux_variables[0], self.aux_variables[0+gradient_offset]]
            grad_h = [self.aux_variables[1], self.aux_variables[1+gradient_offset]]
            dalpha_dx = self.aux_variables[2:2+offset]
            dalpha_dy = self.aux_variables[2+gradient_offset: 2+gradient_offset+offset]
            dU_dx = sum([dalpha_dx[k] * phi[k] for k in range(self.level+1)])
            dU_dy = sum([dalpha_dy[k] * phi[k] for k in range(self.level+1)])
            grad_alpha = [dalpha_dx, dalpha_dy]
            
            dbeta_dx = self.aux_variables[2+offset: 2+2*offset]
            dbeta_dy = self.aux_variables[2+offset+gradient_offset: 2+2*offset+gradient_offset]
            dV_dx = sum([dbeta_dx[k] * phi[k] for k in range(self.level+1)])
            dV_dy = sum([dbeta_dy[k] * phi[k] for k in range(self.level+1)])
            grad_beta = [dbeta_dx, dbeta_dy]
            grad_U = [[dU_dx, dU_dy], [dV_dx, dV_dy]]
        else:
            assert False



        return grad_b, grad_h, grad_alpha, grad_beta, grad_U
    
    def get_Up(self):
        b, h, alpha, beta, hinv = self.get_primitives()
        phi = self.basisfunctions.basis
        z = sympy.Symbol('z')

        # velocity components at bottom (z=0)
        Up = sum(alpha[k] * phi[k].subs(z, 0) for k in range(self.level + 1))
        Vp = 0
        if self.dimension == 2:
            Vp = sum(beta[k] * phi[k].subs(z, 0) for k in range(self.level + 1))
        return [Up, Vp]
    
    def get_ustar(self):
        p = self.parameters
        b, h, alpha, beta, hinv = self.get_primitives()
        phi = self.basisfunctions.basis
        z = sympy.Symbol('z')

        [Up, Vp] = self.get_Up()

        # magnitude
        Umag = sympy.sqrt(Up**2 + Vp**2)
        ReU = (Umag * p.yp) / p.nu

        
        # smooth-wall fallback if Ks ~ 0
        smooth = (p.Ks < 1e-8)

        u_star = sympy.Piecewise(
            # viscous sublayer
            (sympy.sqrt(Umag * p.nu / p.yp), ReU < 30),
            # rough wall (OpenFOAM form)
            (p.kappa * Umag / sympy.log((p.yp * p.CsW) / p.Ks), sympy.Not(smooth)),
            # smooth wall (Lambert W)
            (p.kappa * Umag / sympy.LambertW((p.kappa * Umag * p.yp / p.nu)
                                        * sympy.exp(p.kappa * p.B)), True)
        )


        return u_star
    

    def get_taub(self):
        
        p = self.parameters
        b, h, alpha, beta, hinv = self.get_primitives()
        phi = self.basisfunctions.basis
        z = sympy.Symbol('z')

        [Up, Vp] = self.get_Up()
        Umag = sympy.sqrt(Up**2 + Vp**2)

        u_star = self.get_ustar()

        eps = 1e-12
        # tangential direction (normalized velocity vector)
        t_hat_x = Up / (Umag + eps)
        t_hat_y = Vp / (Umag + eps)

        # shear stress components (vector form)
        tau_zx = - u_star**2 * t_hat_x
        tau_zy = - u_star**2 * t_hat_y
        
        # tau_zx = - 1e-4 * (Up - 0) /0.001
        # tau_zy = - 1e-4 * (Vp - 0) /0.001
        return [tau_zx, tau_zy]
    
    def get_taub_linear_gradient(self):
        
        p = self.parameters
        b, h, alpha, beta, hinv = self.get_primitives()
        phi = self.basisfunctions.basis
        z = sympy.Symbol('z')

        [Up, Vp] = self.get_Up()
        
        tau_zx = - (p.nu + self.lm().subs(z, 0) * self.abs_Sij().subs(z, 0)) * (Up - 0) / p.yp
        tau_zy = - (p.nu + self.lm().subs(z, 0) * self.abs_Sij().subs(z,0)) * (Vp - 0) / p.yp
        return [tau_zx, tau_zy]
    
    
    def get_deltaX(self):
        gradient_offset = self.n_variables
        return self.aux_variables[2*gradient_offset]


    def project_2d_to_3d(self):
        out = Matrix([0 for i in range(6)])
        level = self.level
        offset = level+1
        offset_aux = self.n_aux_variables
        x = self.position[0]
        y = self.position[1]
        z = self.position[2]
        
        b, h, alpha, beta, hinv = self.get_primitives()
        grad_b, grad_h, grad_alpha, grad_beta, grad_U = self.get_gradient()
        
        psi = [self.basisfunctions.eval_psi(k, z) for k in range(level+1)]
        phi = [self.basisfunctions.eval(k, z) for k in range(level+1)]

        rho_w = 1000.
        g = 9.81
        u_3d = self.basismatrices.basisfunctions.reconstruct_velocity_profile_at(alpha, z)
        v_3d = 0
        def dot(a, b):
            s = 0
            for i in range(len(a)):
                s += a[i] * b[i]
            return s
        w_3d = - grad_h[0] * dot(alpha,psi) - h * dot(grad_alpha[0],psi) + dot(alpha, phi) * (z * grad_h[0] + grad_b[0])
        if self.dimension == 2:
            v_3d = self.basismatrices.basisfunctions.reconstruct_velocity_profile_at(beta, z)
            w_3d += - grad_h[1] * dot(beta,psi) - h * dot(grad_beta[1],psi) + dot(beta, phi) * (z * grad_h[1] + grad_b[1])

        out[0] = b
        out[1] = h
        out[2] = u_3d
        out[3] = v_3d
        out[4] = w_3d
        out[5] = rho_w * g * h * (1-z)
        return out

    def flux(self):
        flux_x = Matrix([0 for i in range(self.n_variables)])
        flux_y = Matrix([0 for i in range(self.n_variables)])
        b, h, alpha, beta, hinv = self.get_primitives()
        p = self.parameters
        flux_x[1] = h * alpha[0]
        flux_x[2] = p.g * p.ez * h * h / 2
        for k in range(self.level + 1):
            for i in range(self.level + 1):
                for j in range(self.level + 1):
                    flux_x[k + 2] += (
                        h * alpha[i] * alpha[j]
                        * self.basismatrices.A[k, i, j]
                        / self.basismatrices.M[k, k]
                    )
        if self.dimension == 2:
            offset = self.level + 1
            p = self.parameters
            for k in range(self.level + 1):
                for i in range(self.level + 1):
                    for j in range(self.level + 1):
                        flux_x[1 + k + 1 + offset] += (
                            h * beta[i] * alpha[j]
                            * self.basismatrices.A[k, i, j]
                            / self.basismatrices.M[k, k]
                        )

            flux_y[1] = h * beta[0]
            flux_y[2 + offset] = p.g * p.ez * h * h / 2
            for k in range(self.level + 1):
                for i in range(self.level + 1):
                    for j in range(self.level + 1):
                        flux_y[1 + k + 1] += (
                            h * beta[i] * alpha[j]
                            * self.basismatrices.A[k, i, j]
                            / self.basismatrices.M[k, k]
                        )
            for k in range(self.level + 1):
                for i in range(self.level + 1):
                    for j in range(self.level + 1):
                        flux_y[1 + k + 1 + offset] += (
                            h * beta[i] * beta[j]
                            * self.basismatrices.A[k, i, j]
                            / self.basismatrices.M[k, k]
                        )
        return [flux_x, flux_y][:self.dimension]

    def nonconservative_matrix(self):
        nc_x = Matrix([[0 for i in range(self.n_variables)] for j in range(self.n_variables)])
        nc_y = Matrix([[0 for i in range(self.n_variables)] for j in range(self.n_variables)])
        b, h, alpha, beta, hinv = self.get_primitives()
        p = self.parameters
        um = alpha[0]
        for k in range(1, self.level + 1):
            nc_x[1+k + 1, 1+k + 1] += um
        for k in range(self.level + 1):
            for i in range(1, self.level + 1):
                for j in range(1, self.level + 1):
                    nc_x[1+k + 1, 1+i + 1] -= (
                        alpha[j]
                        * self.basismatrices.B[k, i, j]
                        / self.basismatrices.M[k, k]
                    )
        if self.dimension == 2:
            offset = self.level + 1
            b, h, alpha, beta, hinv = self.get_primitives()
            p = self.parameters
            um = alpha[0]
            vm = beta[0]
            for k in range(1, self.level + 1):
                nc_y[1+k + 1, 1+k + 1 + offset] += um
            for k in range(self.level + 1):
                for i in range(1, self.level + 1):
                    for j in range(1, self.level + 1):
                        nc_x[1+k + 1, 1+i + 1] -= (
                            alpha[j]
                            * self.basismatrices.B[k, i, j]
                            / self.basismatrices.M[k, k]
                        )
                        nc_y[1+k + 1, 1+i + 1 + offset] -= (
                            alpha[j]
                            * self.basismatrices.B[k, i, j]
                            / self.basismatrices.M[k, k]
                        )

            for k in range(1, self.level + 1):
                nc_x[1+k + 1 + offset, 1+k + 1] += vm
                nc_y[1+k + 1 + offset, 1+k + 1 + offset] += vm
            for k in range(self.level + 1):
                for i in range(1, self.level + 1):
                    for j in range(1, self.level + 1):
                        nc_x[1+k + 1 + offset, 1+i + 1] -= (
                            beta[j]
                            * self.basismatrices.B[k, i, j]
                            / self.basismatrices.M[k, k]
                        )
                        nc_y[1+k + 1 + offset, 1+i + 1 + offset] -= (
                            beta[j]
                            * self.basismatrices.B[k, i, j]
                            / self.basismatrices.M[k, k]
                        )
        return [-nc_x, -nc_y][:self.dimension]

    def eigenvalues(self):
        # we delete heigher order moments (level >= 2) for analytical eigenvalues
        offset = self.level + 1
        A = self.normal[0] * self.quasilinear_matrix()[0]
        for d in range(1, self.dimension):
            A += self.normal[d] * self.quasilinear_matrix()[d]
        b, h, alpha, beta, hinv = self.get_primitives()
        alpha_erase = alpha[1:] if self.level >= 2 else []
        beta_erase = beta[1:] if self.level >= 2 else []
        for alpha_i in alpha_erase:
            A = A.subs(alpha_i, 0)
        for beta_i in beta_erase:
            A = A.subs(beta_i, 0)
        return eigenvalue_dict_to_matrix(A.eigenvals())
    

    def Sij(self):
        out = sympy.zeros(self.dimension, self.dimension)

        grad_b, grad_h, grad_alpha, grad_beta, grad_U = self.get_gradient()
        for d1 in range(self.dimension):
            for d2 in range(self.dimension):
                out[d1, d2] = 0.5 * (grad_U[d1][d2] + grad_U[d2][d1])

        return out
    
    def abs_Sij(self):
        Sij = self.Sij()
        out = 0
        for i in range(self.dimension):
            for j in range(self.dimension):
                out += Sij[i,j]**2
        out = sympy.sqrt(2 * out)
        return out
    
    def Szj(self):
        out = sympy.zeros(self.dimension)

        b, h, alpha, beta, hinv = self.get_primitives()
        dphi_dz = self.basisfunctions.get_diff_basis()
        
        for k in range(self.level+1):
            out[0] += alpha[k] * dphi_dz[k]
        if self.dimension == 2:
            for k in range(self.level+1):
                out[1] += beta[k] * dphi_dz[k]
        return out
    
    def source(self):
        out = Matrix([0 for i in range(self.n_variables)])
        out += self.mixing_length_vertical()
        # out += self.mixing_length_vertical_linear_gradient()
        out += self.smagorinsky_source()
        return out
    
    def source_implicit(self):
        out = sympy.zeros(self.n_variables, self.n_variables)
        # out += self.taub_implicit()
        return out
    
    def taub_implicit(self):
        out = sympy.zeros(self.n_variables, self.n_variables)
        b, h, alpha, beta, hinv = self.get_primitives()
        phi = self.basisfunctions.basis
        z = sympy.Symbol('z')
        u_star = self.get_ustar()
        Up = self.get_Up()
        magU = sympy.sqrt(Up[0]**2 + Up[1]**2)
        eps = 10**(-8)
        for d in range(self.dimension):
            term = u_star**2 / h /(magU+eps) 
            for k in range(self.level+1):
                for i in range(self.level+1):
                    ik = 2 + k + d * (self.level + 1)
                    ii = 2 + i + d * (self.level + 1)
                    out[ik, ii] +=  term * phi[k].subs(z, 0) * phi[i].subs(z, 0) / self.basismatrices.M[k, k]
        return out
    
    def dflux(self):
        return self.smagorinsky_dflux()
    
    def smagorinsky_dflux(self):
        """
        diffusive flux due to the Smagorinsky model
        """
        assert "Cs" in vars(self.parameters)

        dflux = [sympy.zeros(self.n_variables, 1) for d in range(self.dimension)]
        z = sympy.Symbol('z')
        level = self.level
        offset = level+1
        phi = [self.basisfunctions.eval(k, z) for k in range(level+1)]
        dX = self.get_deltaX()
        Cs = self.parameters.Cs
        
        xi, wi = gauss_legendre(self.order_numerical_integration, 8)
        xi = [0.5 * (x + 1) for x in xi]
        wi = [0.5 * w for w in wi]
        b, h, alpha, beta, hinv = self.get_primitives()

        abs_Sij = self.abs_Sij()
        Sij = self.Sij()
        
        for i in range(len(xi)):
            zi = xi[i]
            for k in range(level+1):
                for d1 in range(self.dimension):
                    for d2 in range(self.dimension):
                        dflux[d1][2 + d2 * offset + k, 0] += wi[i] * h*  phi[k].subs(z, zi) * 2 * (Cs * dX)**2 * abs_Sij.subs(z, zi) * Sij[d1, d2].subs(z, zi)
        for k in range(level+1):
            for d in range(self.dimension):
                dflux[d][2+k, 0] /= self.basismatrices.M[k, k]
        return [dflux[d] for d in range(self.dimension)]
    


        
    def smagorinsky_source(self):
        out = Matrix([0 for i in range(self.n_variables)])
        z = sympy.Symbol('z')
        level = self.level
        offset = level+1
        phi = self.basisfunctions.basis
        dphi_dz = self.basisfunctions.get_diff_basis()
        dX = self.get_deltaX()
        Cs = self.parameters.Cs
        
        xi, wi = gauss_legendre(self.order_numerical_integration, 8)
        xi = [0.5 * (x + 1) for x in xi]
        wi = [0.5 * w for w in wi]
        b, h, alpha, beta, hinv = self.get_primitives()
        grad_b, grad_h, grad_alpha, grad_beta, grad_U = self.get_gradient()
        sigma = [grad_b[0] + z * grad_h[0], grad_b[1] + z * grad_h[1]]
        
        
        abs_Sij = self.abs_Sij()
        Sij = self.Sij()
        
        taub = self.get_taub()
        for d in range(self.dimension):
            for k in range(1 + self.level):
                out[2+k + d*offset] += sigma[d].subs(z, 0) * taub[d] * phi[k].subs(z, 0) / self.basismatrices.M[k, k]
            
        
        for i in range(len(xi)):
            zi = xi[i]
            for k in range(level+1):
                for d1 in range(self.dimension):
                    for d2 in range(self.dimension):
                        out[2 + d2 * offset + k] -= wi[i] * sigma[d1].subs(z, zi) *  2 * (Cs * dX)**2 * abs_Sij.subs(z, zi) * Sij[d1, d2].subs(z, zi)  * dphi_dz[k].subs(z, zi) / self.basismatrices.M[k, k]
        return out

    def lm(self):
        p = self.parameters
        b, h, alpha, beta, hinv = self.get_primitives()
        z = sympy.Symbol('z')
        return p.kappa * (z + p.yp) * (1 - z)
    
    def mixing_length_numerical_term(self):
        assert "kappa" in vars(self.parameters)
        out = 0
        b, h, alpha, beta, hinv = self.get_primitives()
        z = sympy.Symbol('z')
        phi = [self.basisfunctions.eval(k, z) for k in range(self.level+1)]
        u = sum([alpha[k] * phi[k] for k in range(self.level+1)])
        v = sum([beta[k] * phi[k] for k in range(self.level+1)] )
        U = [u,v][:self.dimension]
        
        for d1 in range(self.dimension):
            for d2 in range(self.dimension):
                out += U[d1] * U[d2]
                
        eps = 1e-10
        out += eps
        out = sympy.sqrt(out)
        out *= (self.lm())**2
        return out


    def mixing_length_vertical(self):
        out = Matrix([0 for i in range(self.n_variables)])
        offset = self.level + 1
        b, h, alpha, beta, hinv = self.get_primitives()
        p = self.parameters
        z = sympy.Symbol('z')
        phi = [self.basisfunctions.eval(k, z) for k in range(self.level+1)]
        dphi_dz = self.basisfunctions.get_diff_basis()
        
        Szj = self.Szj()

        taub = self.get_taub()
        for d in range(self.dimension):
            for k in range(1 + self.level):
                out[2+k + d*offset] += taub[d] * phi[k].subs(z, 0) / self.basismatrices.M[k, k]
            
        xi, wi = gauss_legendre(self.order_numerical_integration, 8)
        xi = [0.5 * (x + 1) for x in xi]
        wi = [0.5 * w for w in wi]
        
        for d in range(self.dimension):
            for i in range(len(xi)):
                for k in range(self.level+1):
                    out[2+k+d*offset] -= wi[i] * self.mixing_length_numerical_term().subs(z, xi[i]) * Szj[d].subs(z, xi[i]) * dphi_dz[k].subs(z, xi[i]) / self.basismatrices.M[k, k]
                    
        return out
    
    def mixing_length_vertical_linear_gradient(self):
        out = Matrix([0 for i in range(self.n_variables)])
        offset = self.level + 1
        b, h, alpha, beta, hinv = self.get_primitives()
        p = self.parameters
        z = sympy.Symbol('z')
        phi = [self.basisfunctions.eval(k, z) for k in range(self.level+1)]
        dphi_dz = self.basisfunctions.get_diff_basis()
        
        Szj = self.Szj()

        taub = self.get_taub_linear_gradient()
        for d in range(self.dimension):
            for k in range(1 + self.level):
                out[2+k + d*offset] += taub[d] * phi[k].subs(z, 0) / self.basismatrices.M[k, k]
            
        xi, wi = gauss_legendre(self.order_numerical_integration, 8)
        xi = [0.5 * (x + 1) for x in xi]
        wi = [0.5 * w for w in wi]
        
        for d in range(self.dimension):
            for i in range(len(xi)):
                for k in range(self.level+1):
                    out[2+k+d*offset] -= wi[i] * self.mixing_length_numerical_term().subs(z, xi[i]) * Szj[d].subs(z, xi[i]) * dphi_dz[k].subs(z, xi[i]) / self.basismatrices.M[k, k]
                    
        return out
    


@define(frozen=True, slots=True, kw_only=True)
class SMETNum(SMET):
    ref_model: Model = field(init=False)
    
    def __attrs_post_init__(self):
        super().__attrs_post_init__()
        object.__setattr__(self, "ref_model", SMET(level=self.level, dimension=self.dimension, boundary_conditions=self.boundary_conditions))

    def flux(self):
        return [self.substitute_precomputed_denominator(f, self.variables[1], self.aux_variables.hinv) for f in self.ref_model.flux()]      
    
    def nonconservative_matrix(self):
        return [self.substitute_precomputed_denominator(f, self.variables[1], self.aux_variables.hinv) for f in self.ref_model.nonconservative_matrix()]  
    
    def quasilinear_matrix(self):
        return [self.substitute_precomputed_denominator(f, self.variables[1], self.aux_variables.hinv) for f in self.ref_model.quasilinear_matrix()]    
    
    def source(self):
        return self.substitute_precomputed_denominator(self.ref_model.source(), self.variables[1], self.aux_variables.hinv)
    
    def source_implicit(self):
        return self.substitute_precomputed_denominator(self.ref_model.source_implicit(), self.variables[1], self.aux_variables.hinv)
    
    def residual(self):
        return self.substitute_precomputed_denominator(self.ref_model.residual(), self.variables[1], self.aux_variables.hinv)
    
    def left_eigenvectors(self):
        return self.substitute_precomputed_denominator(self.ref_model.left_eigenvectors(), self.variables[1], self.aux_variables.hinv)
    
    def right_eigenvectors(self):
        return self.substitute_precomputed_denominator(self.ref_model.right_eigenvectors(), self.variables[1], self.aux_variables.hinv)

    def eigenvalues(self):
        h = self.variables[1]
        evs = self.substitute_precomputed_denominator(self.ref_model.eigenvalues(), self.variables[1], self.aux_variables.hinv)
        for i in range(self.n_variables):
            evs[i] = Piecewise((evs[i], h > 1e-8), (0, True))
        return evs



import numpy as np
from time import time as get_time


import sympy
from sympy import Matrix

from attr import define, field
from typing import Callable, List

from zoomy_core.misc.misc import Zstruct, ZArray

from zoomy_core.model.basefunction import Function


@define(slots=True, frozen=False, kw_only=True)
class BoundaryCondition:
    tag: str

    """ 
    Default implementation. The required data for the 'ghost cell' is the data from the interior cell. Can be overwritten e.g. to implement periodic boundary conditions.
    """

    def compute_boundary_condition(self, time, X, dX, Q, Qaux, parameters, normal):
        print("BoundaryCondition is a virtual class. Use one if its derived classes!")
        assert False


@define(slots=True, frozen=False, kw_only=True)
class Extrapolation(BoundaryCondition):
    def compute_boundary_condition(self, time, X, dX, Q, Qaux, parameters, normal):
        return ZArray(Q)


@define(slots=True, frozen=False, kw_only=True)
class InflowOutflow(BoundaryCondition):
    prescribe_fields: dict[int, float]

    def compute_boundary_condition(self, time, X, dX, Q, Qaux, parameters, normal):
        Qout = ZArray(Q)
        for k, v in self.prescribe_fields.items():
            Qout[k] = eval(v)
        return Qout


@define(slots=True, frozen=False, kw_only=True)
class Lambda(BoundaryCondition):
    prescribe_fields: dict[int, float]

    def compute_boundary_condition(self, time, X, dX, Q, Qaux, parameters, normal):
        Qout = ZArray(Q)
        for k, v in self.prescribe_fields.items():
            Qout[k] = v(time, X, dX, Q, Qaux, parameters, normal)
        return Qout


def _sympy_interpolate_data(time, timeline, data):
    assert timeline.shape[0] == data.shape[0]
    conditions = (((data[0], time <= timeline[0])),)
    for i in range(timeline.shape[0] - 1):
        t0 = timeline[i]
        t1 = timeline[i + 1]
        y0 = data[i]
        y1 = data[i + 1]
        conditions += (
            (-(time - t1) / (t1 - t0) * y0 + (time - t0) / (t1 - t0) * y1, time <= t1),
        )
    conditions += (((data[-1], time > timeline[-1])),)
    return sympy.Piecewise(*conditions)


@define(slots=True, frozen=False, kw_only=True)
class FromData(BoundaryCondition):
    prescribe_fields: dict[int, np.ndarray]
    timeline: np.ndarray

    def compute_boundary_condition(self, time, X, dX, Q, Qaux, parameters, normal):
        # Extrapolate all fields
        Qout = ZArray(Q)

        # Set the fields which are prescribed in boundary condition dict
        time_start = get_time()
        for k, v in self.prescribe_fields.items():
            interp_func = _sympy_interpolate_data(time, self.timeline, v)
            Qout[k] = 2 * interp_func - Q[k]
        return Qout


@define(slots=True, frozen=False, kw_only=True)
class Wall(BoundaryCondition):
    """
    momentum_field_indices: list(int): indicate which fields need to be mirrored at the wall
    permeability: float : 1.0 corresponds to a perfect reflection (impermeable wall)
    """

    momentum_field_indices: List[List[int]] = [[1, 2]]
    permeability: float = 0.0
    wall_slip: float = 1.0

    def compute_boundary_condition(self, time, X, dX, Q, Qaux, parameters, normal):
        q = Matrix(Q.get_list())
        n_variables = Q.length()
        momentum_list = [Matrix([q[k] for k in l]) for l in self.momentum_field_indices]
        dim = momentum_list[0].shape[0]
        n = Matrix(normal[:dim])
        out = Matrix.zeros(n_variables)
        out = q
        momentum_list_wall = []
        for momentum in momentum_list:
            normal_momentum_coef = momentum.dot(n)
            transverse_momentum = momentum - normal_momentum_coef * n
            momentum_wall = (
                self.wall_slip * transverse_momentum
                - (1 - self.permeability) * normal_momentum_coef * n
            )
            momentum_list_wall.append(momentum_wall)
        for l, momentum_wall in zip(self.momentum_field_indices, momentum_list_wall):
            for i_k, k in enumerate(l):
                out[k] = momentum_wall[i_k]
        return out


@define(slots=True, frozen=False, kw_only=True)
class RoughWall(Wall):
    CsW: float = 0.5  # roughness constant
    Ks: float = 0.001  # roughness height

    def compute_boundary_condition(self, time, X, dX, Q, Qaux, parameters, normal):
        slip_length = dX * sympy.ln((dX * self.CsW) / self.Ks)
        # wall_slip = (1-2*dX/slip_length)
        # wall_slip = sympy.Min(sympy.Max(wall_slip, 0.0), 1.0)
        f = dX / slip_length
        wall_slip = (1 - f) / (1 + f)
        self.wall_slip = wall_slip
        return super().compute_boundary_condition(
            time, X, dX, Q, Qaux, parameters, normal
        )


@define(slots=True, frozen=False, kw_only=True)
class Periodic(BoundaryCondition):
    periodic_to_physical_tag: str

    def compute_boundary_condition(self, time, X, dX, Q, Qaux, parameters, normal):
        return ZArray(Q)


@define(slots=True, frozen=True)
class BoundaryConditions:
    _boundary_conditions: List[BoundaryCondition]
    _boundary_functions: List[Callable] = field(init=False)
    _boundary_tags: List[str] = field(init=False)

    def __attrs_post_init__(self):
        tags_unsorted = [bc.tag for bc in self._boundary_conditions]
        order = np.argsort(tags_unsorted)
        object.__setattr__(
            self,
            "_boundary_functions",
            [self._boundary_conditions[i].compute_boundary_condition for i in order],
        )
        object.__setattr__(
            self,
            "_boundary_tags",
            [self._boundary_conditions[i].tag for i in order],
        )
        object.__setattr__(
            self, "_boundary_conditions", [self._boundary_conditions[i] for i in order]
        )

    def get_boundary_condition_function(self, time, X, dX, Q, Qaux, parameters, normal):
        bc_idx = sympy.Symbol("bc_idx", integer=True)
        bc_func = sympy.Piecewise(
            *(
                (func(time, X.get_list(), dX, Q.get_list(), Qaux.get_list(), parameters.get_list(), normal.get_list()), sympy.Eq(bc_idx, i))
                for i, func in enumerate(self._boundary_functions)
            )
        )
        func = Function(
            name="boundary_conditions",
            args=Zstruct(
                idx=bc_idx,
                time=time,
                position=X,
                distance=dX,
                variables=Q,
                aux_variables=Qaux,
                parameters=parameters,
                normal=normal,
            ),
            definition=bc_func,
        )
        return func



import os

try:
    import h5py

    _HAVE_H5PY = True
except ImportError:
    _HAVE_H5PY = False


try:
    from petsc4py import PETSc

    _HAVE_PETSC = True
except ImportError:
    _HAVE_PETSC = False


try:
    import meshio

    _HAVE_MESHIO = True
except:
    _HAVE_MESHIO = False
from copy import deepcopy
from itertools import product
from typing import Union

import numpy as np
from attr import define

import zoomy_core.mesh.mesh_extrude as extrude
import zoomy_core.mesh.mesh_util as mesh_util
from zoomy_core.mesh.mesh_util import compute_subvolume, get_extruded_mesh_type
from zoomy_core.misc.custom_types import CArray, FArray, IArray
from zoomy_core.misc.static_class import register_static_pytree
from zoomy_core.model.boundary_conditions import Periodic

# petsc4py.init(sys.argv)


def build_monomial_indices(degree: int, dim: int):
    """
    Build the full set of monomial multi-indices for a given degree and dimension.

    Parameters:
    - degree (int): Maximum total polynomial degree
    - dim (int): Number of dimensions

    Returns:
    - List of tuples: each tuple is a multi-index (α₁, ..., α_dim)
    """
    mon_indices = []
    for powers in product(range(degree + 1), repeat=dim):
        if sum(powers) <= degree:
            if sum(powers) == 0:
                continue
            mon_indices.append(powers)
    return mon_indices


def scale_lsq_derivative(mon_indices):
    from math import factorial

    import numpy as np

    scale_factors = np.array(
        [np.prod([factorial(k) for k in mi]) for mi in mon_indices]
    )  # shape (n_monomials,)
    return scale_factors


def find_derivative_indices(full_monomials_arr, requested_derivs_arr):
    """
    Args:
        full_monomials_arr: shape (N, dim)
        requested_derivs_arr: shape (M, dim)

    Returns:
        indices: shape (M,), where indices[i] gives the index of requested_derivs_arr[i]
                 in full_monomials_arr, or -1 if not found
    """
    # Ensure arrays
    full_monomials_arr = np.array(full_monomials_arr)
    requested_derivs_arr = np.array(requested_derivs_arr)

    # Compare all requested derivatives with all monomials
    matches = np.all(
        full_monomials_arr[:, None, :] == requested_derivs_arr[None, :, :], axis=-1
    )  # (N, M)

    # Which monomials match each requested derivative?
    found = np.any(matches, axis=0)  # (M,)
    indices = np.argmax(matches, axis=0)  # (M,) — returns 0 even if not found

    # Replace unfound indices with -1
    indices = np.where(found, indices, -1)
    return indices


def compute_derivatives(u, mesh, derivatives_multi_index=None):
    A_glob = mesh.lsq_gradQ  # shape (n_cells, n_monomials, n_neighbors)
    neighbors = mesh.lsq_neighbors  # list of neighbors per cell
    mon_indices = mesh.lsq_monomial_multi_index  # shape (n_monomials, dim)
    # scale_factors = scale_lsq_derivative(mon_indices)
    scale_factors = mesh.lsq_scale_factors

    if derivatives_multi_index is None:
        derivatives_multi_index = mon_indices
    indices = find_derivative_indices(mon_indices, derivatives_multi_index)

    def reconstruct_cell(A_loc, neighbor_idx, u_i):
        u_neighbors = u[neighbor_idx]
        delta_u = u_neighbors - u_i
        return (scale_factors * (A_loc.T @ delta_u)).T  # shape (n_monomials,)

    out = np.zeros((A_glob.shape[0], len(
        derivatives_multi_index)), dtype=float)
    for i in range(A_glob.shape[0]):
        A_loc = A_glob[i]
        neighbor_idx = neighbors[i]
        u_i = u[i]
        out[i, :] = reconstruct_cell(A_loc, neighbor_idx, u_i)

    return out


def get_polynomial_degree(mon_indices):
    return max(sum(m) for m in mon_indices)


def get_required_monomials_count(degree, dim):
    from math import comb

    return comb(int(degree + dim), int(dim))


def build_vandermonde(cell_diffs, mon_indices):
    n_neighbors, dim = cell_diffs.shape
    n_monomials = len(mon_indices)
    V = np.zeros((n_neighbors, n_monomials))
    for i, powers in enumerate(mon_indices):
        V[:, i] = np.prod(cell_diffs**powers, axis=1)
    return V


def expand_neighbors(neighbors_list, initial_neighbors):
    # Expand to the next ring: union of neighbors of neighbors
    expanded = set(initial_neighbors)
    for n in initial_neighbors:
        expanded.update(neighbors_list[n])
    return list(expanded)


def compute_gaussian_weights(dX, sigma=1.0):
    distances = np.linalg.norm(dX, axis=1)
    weights = np.exp(-((distances / sigma) ** 2))
    return weights


def least_squares_reconstruction_local(
    n_cells, dim, neighbors_list, cell_centers, lsq_degree
):
    """
    Build the full set of monomial multi-indices for a given degree and dimension.

    Parameters:
    - degree (int): Maximum total polynomial degree
    - dim (int): Number of dimensions

    Returns:
    - List of tuples: each tuple is a multi-index (α₁, ..., α_dim)
    """
    mon_indices = build_monomial_indices(lsq_degree, dim)
    A_glob = []
    neighbors_all = []  # to store expanded neighbors per cell
    degree = get_polynomial_degree(mon_indices)
    required_neighbors = get_required_monomials_count(degree, dim)

    # First, expand neighborhoods and store them
    for i_c in range(n_cells):
        current_neighbors = list(neighbors_list[i_c])

        # Expand neighbor rings until we have enough
        while len(current_neighbors) < required_neighbors:
            new_neighbors = expand_neighbors(neighbors_list, current_neighbors)
            current_neighbors = list(set(new_neighbors) - {i_c})

        neighbors_all.append(current_neighbors)

    # Compute max neighbors after expansion
    max_neighbors = max(len(nbrs) for nbrs in neighbors_all)

    A_glob = []
    neighbors_array = np.empty((n_cells, max_neighbors), dtype=int)

    for i_c in range(n_cells):
        current_neighbors = list(neighbors_all[i_c])  # previously expanded
        n_nbr = len(current_neighbors)

        # Expand further if still under-resolved
        while n_nbr < max_neighbors:
            extended_neighbors = expand_neighbors(
                neighbors_list, current_neighbors)
            extended_neighbors = list(
                set(extended_neighbors) - {i_c})  # Remove self
            # Keep only new neighbors not already present
            new_neighbors = [
                n for n in extended_neighbors if n not in current_neighbors
            ]
            current_neighbors.extend(new_neighbors)
            n_nbr = len(current_neighbors)

            # Stop if enough neighbors collected
            if n_nbr >= max_neighbors:
                break

        # Trim or pad neighbor list to exact max_neighbors
        if len(current_neighbors) >= max_neighbors:
            trimmed_neighbors = current_neighbors[:max_neighbors]
        else:
            padding = [i_c] * (max_neighbors - len(current_neighbors))
            trimmed_neighbors = current_neighbors + padding

        neighbors_array[i_c, :] = trimmed_neighbors

        # Build dX matrix for least squares
        dX = np.zeros((max_neighbors, dim), dtype=float)
        for j, neighbor in enumerate(trimmed_neighbors):
            dX[j, :] = cell_centers[neighbor] - cell_centers[i_c]

        # shape (max_neighbors, n_monomials)
        V = build_vandermonde(dX, mon_indices)

        # Compute weights
        weights = compute_gaussian_weights(dX)  # shape (n_neighbors,)
        W = np.diag(weights)  # shape (n_neighbors, n_neighbors)
        VW = W @ V  # shape (n_neighbors, n_monomials)
        alpha = dX.min() * 10 ** (-8)
        A_loc = (
            np.linalg.pinv(VW.T @ VW + alpha * np.eye(VW.shape[1])) @ VW.T @ W
        )  # shape (n_monomials, n_neighbors)
        # A_loc = np.linalg.pinv(V).T  # shape (n_monomials, max_neighbors)
        A_glob.append(A_loc.T)

    A_glob = np.array(A_glob)  # shape (n_cells, n_monomials, max_neighbors)
    return A_glob, neighbors_array, mon_indices


def get_physical_boundary_labels(filepath):
    if not _HAVE_MESHIO:
        raise RuntimeError(
            "_write_to_vtk_from_vertices_edges requires meshio, which is not available."
        )
    mesh = meshio.read(filepath)
    boundary_dict = {key: value[0] for key, value in mesh.field_data.items()}
    return boundary_dict


def _compute_inradius_generic(cell_center, face_centers, face_normals):
    """
    strategy: find the shortest path from the center to each side (defined by face center and normal)
    use the minimum of all these shortest paths
    For a distorted element, the inradius might be zero. In this case, use minimal distance of center to face_centers
    """
    inradius = np.inf
    for center, normal in zip(face_centers, face_normals):
        distance = np.abs(np.dot(center - cell_center, normal))
        inradius = min(inradius, distance)
    if inradius <= 0:
        inradius = np.array(face_centers - cell_center).min()
    return inradius


def compute_cell_inradius(dm):
    """
    Error in petsc4py? dm.getMinRadius() always returns 0
    strategy: find the shortest path from the center to each side (defined by face center and normal)
    use the minimum of all these shortest paths
    """
    (cStart, cEnd) = dm.getHeightStratum(0)
    (vStart, vEnd) = dm.getDepthStratum(0)
    (eStart, eEnd) = dm.getDepthStratum(1)
    inradius = []
    for c in range(cStart, cEnd):
        _, cell_center, _ = dm.computeCellGeometryFVM(c)
        face_normals = []
        face_centers = []
        faces = dm.getCone(c)
        for f in faces:
            _, center, normal = dm.computeCellGeometryFVM(f)
            face_normals.append(normal)
            face_centers.append(center)
        inradius.append(
            _compute_inradius_generic(cell_center, face_centers, face_normals)
        )
    return np.array(inradius, dtype=float)


def get_mesh_type_from_dm(num_faces_per_cell, dim):
    if dim == 1:
        if num_faces_per_cell == 2:
            return "line"
        else:
            assert False
    elif dim == 2:
        if num_faces_per_cell == 3:
            return "triangle"
        elif num_faces_per_cell == 4:
            return "quad"
        else:
            assert False
    elif dim == 3:
        if num_faces_per_cell == 4:
            return "tetra"
        elif num_faces_per_cell == 6:
            return "edge"
        elif num_faces_per_cell == 8:
            return "hexahedron"
        else:
            assert False
    assert False


def _boundary_dict_to_list(d):
    l = []
    for i, vs in enumerate(d.values()):
        l += vs
    return l


def _boundary_dict_indices(d):
    indices = []
    index = 0
    for values in d.values():
        indices += [index for v in values]
        index += 1
    return np.array(indices, dtype=int)


def _get_neighberhood(dm, cell, cStart=0):
    neighbors = (
        np.array(
            [dm.getSupport(f)[dm.getSupport(f) != cell][0]
             for f in dm.getCone(cell)],
            dtype=int,
        )
        - cStart
    )
    return neighbors


def _fill_neighborhood(dm, neighbors, max_neighbors, cStart=0):
    new_potential_neighbors = []
    for cell in neighbors:
        new_potential_neighbors += list(_get_neighberhood(dm,
                                        cell, cStart=cStart))
    new_potential_neighbors = np.array(new_potential_neighbors)
    new_neighbors = np.setdiff1d(new_potential_neighbors, neighbors)
    neighbors = np.concatenate((neighbors, new_neighbors))[:max_neighbors]
    if len(neighbors) == max_neighbors:
        return neighbors
    else:
        return _fill_neighborhood(dm, neighbors, max_neighbors, cStart=cStart)


@register_static_pytree
@define(slots=True, frozen=True)
class Mesh:
    dimension: int
    type: str
    n_cells: int
    n_inner_cells: int
    n_faces: int
    n_vertices: int
    n_boundary_faces: int
    n_faces_per_cell: int
    vertex_coordinates: FArray
    cell_vertices: IArray
    cell_faces: IArray
    cell_volumes: FArray
    cell_centers: FArray
    cell_inradius: FArray
    cell_neighbors: IArray
    boundary_face_cells: IArray
    boundary_face_ghosts: IArray
    boundary_face_function_numbers: IArray
    boundary_face_physical_tags: IArray
    boundary_face_face_indices: IArray
    face_cells: IArray
    # face_cell_face_index: IArray
    face_normals: FArray
    face_volumes: FArray
    face_centers: FArray
    face_subvolumes: FArray
    face_neighbors: IArray
    boundary_conditions_sorted_physical_tags: IArray
    boundary_conditions_sorted_names: CArray
    lsq_gradQ: FArray
    lsq_neighbors: FArray
    lsq_monomial_multi_index: int
    lsq_scale_factors: FArray
    z_ordering: IArray

    def resolve_periodic_bcs(self, bcs):
        """
        Goal: if 'apply_boundary_condition' is called, the ghost cell value is computed, given an input cell value funtion.
              In case of a periodic BC, this is NOT the adjacent cell. So I need to change the 'boundary_face_cell' for the periodic
              cells to point to the right data location!
              This is why we only overwrite the 'mesh.boundary_face_cell' in the end.
              Furthermore, I CANNOT alter any ordering! However, I need to sort the two boundaries such that the e.g.
              left and right border are mapped correctly, as the boundary cells are not ordered.
              As ghost/inner cells is confusing here, I rather like to use 'from' and 'to', as 'from' data from which bc is copied and 'to' stands for the boundary where it is copied to. As we only change the cells where the data is taken from (from the adjacent to the periodic cell), we only alter the 'boundary_face_cell' in the end.
        """
        dict_physical_name_to_index = {
            v: i for i, v in enumerate(self.boundary_conditions_sorted_names)
        }
        dict_function_index_to_physical_tag = {
            i: v for i, v in enumerate(self.boundary_conditions_sorted_physical_tags)
        }

        boundary_face_cells_copy = deepcopy(self.boundary_face_cells)

        for i_bc, bc in enumerate(bcs._boundary_conditions):
            if type(bc) == Periodic:
                from_physical_tag = dict_function_index_to_physical_tag[
                    dict_physical_name_to_index[bc.periodic_to_physical_tag]
                ]
                to_physical_tag = dict_function_index_to_physical_tag[
                    dict_physical_name_to_index[bc.tag]
                ]

                mask_face_from = self.boundary_face_physical_tags == from_physical_tag
                mask_face_to = self.boundary_face_physical_tags == to_physical_tag

                # I want to copy from boundary_face_cells to boundary_face_ghosts!
                from_cells = self.boundary_face_cells[mask_face_from]
                to_cells = self.boundary_face_ghosts[mask_face_to]

                from_coords = self.cell_centers[:, from_cells]
                to_coords = self.cell_centers[:, to_cells]

                # sort not dimension by dimension, but most significant dimension to least significant dimension
                # determine significance by max difference
                significance_per_dimension = [
                    from_coords[d, :].max() - from_coords[d, :].min()
                    for d in range(self.dimension)
                ]
                _significance_per_dimension = [
                    to_coords[d, :].max() - to_coords[d, :].min()
                    for d in range(self.dimension)
                ]
                # reverse the order of lexsort such that the most important is first IS NOT NEEDED, since lexsort starts sorting by the last entry in the list
                sort_order_significance = np.lexsort(
                    [significance_per_dimension])

                from_cells_sort_order = np.lexsort(
                    [from_coords[d, :] for d in sort_order_significance]
                )
                to_cells_sort_order = np.lexsort(
                    [to_coords[d, :] for d in sort_order_significance]
                )

                # generates indices from 0 to number of ghost_cells (total)
                indices = np.array(list(range(mask_face_to.shape[0])))
                # masks away all cells that do not belong to this tag
                indices_to = indices[mask_face_to]
                # sort the indices
                indices_to_sort = indices_to[to_cells_sort_order]

                indices_from = indices[mask_face_from]
                indices_from_sort = indices_from[from_cells_sort_order]

                self.boundary_face_cells[indices_to_sort] = boundary_face_cells_copy[
                    indices_from_sort
                ]

    @classmethod
    def create_1d(cls, domain: tuple[float, float], n_inner_cells: int, lsq_degree=1):
        xL = domain[0]
        xR = domain[1]

        n_cells = n_inner_cells + 2
        n_vertices = n_inner_cells + 1
        dimension = 1
        n_faces_per_cell = 2
        n_faces = n_inner_cells + 1
        n_boundary_faces = 2
        dx = (xR - xL) / n_inner_cells
        vertex_coordinates = np.zeros((n_vertices, 1))
        vertex_coordinates[:, 0] = np.linspace(xL, xR, n_vertices, dtype=float)
        cell_vertices = np.zeros((n_inner_cells, n_faces_per_cell), dtype=int)
        cell_vertices[:, 0] = list(range(0, n_vertices - 1))
        cell_vertices[:, 1] = list(range(1, n_vertices))
        cell_volumes = dx * np.ones(n_cells, dtype=float)
        cell_inradius = dx / 2 * np.ones(n_cells, dtype=float)

        face_normals = np.zeros((n_faces, dimension), dtype=float)
        face_volumes = np.ones((n_faces), dtype=float)

        cell_centers = np.zeros((n_cells, 3), dtype=float)
        cell_centers[:n_inner_cells, 0] = np.arange(xL + dx / 2, xR, dx)
        cell_centers[n_inner_cells, 0] = xL - dx / 2
        cell_centers[n_inner_cells + 1, 0] = xR + dx / 2
        cell_neighbors = (n_cells + 1) * \
            np.ones((n_cells, n_faces_per_cell), dtype=int)

        cell_faces = np.empty((n_inner_cells, n_faces_per_cell), dtype=int)
        cell_faces[:, 0] = list(range(0, n_faces - 1))
        cell_faces[:, 1] = list(range(1, n_faces))

        # inner cells
        for i_cell in range(0, n_cells):
            cell_neighbors[i_cell, :] = [i_cell - 1, i_cell + 1]
        # left neighbor of 0is the first ghost
        cell_neighbors[0, 0] = n_inner_cells
        # right neighbor of n_inner_cell is the second ghost
        cell_neighbors[n_inner_cells - 1, 1] = n_inner_cells + 1
        # left neighbor of first ghost is empty, but we add the neighbor of the neighbor
        cell_neighbors[n_inner_cells, 0] = 1
        # right neighbor of first ghost is first cell
        cell_neighbors[n_inner_cells, 1] = 0
        # left neighbor of second ghost is last inner
        cell_neighbors[n_inner_cells + 1, 0] = n_inner_cells - 1
        # right neighbor of second ghost is empty, but we add the neighbor of the neighbor
        cell_neighbors[n_inner_cells + 1, 1] = n_inner_cells - 2

        for i_face in range(0, n_faces):
            face_normals[i_face, 0] = 1.0

        boundary_face_cells = np.array([0, n_inner_cells - 1], dtype=int)
        boundary_face_ghosts = np.array(
            [n_inner_cells, n_inner_cells + 1], dtype=int)
        boundary_face_function_numbers = np.empty(
            (n_boundary_faces), dtype=int)
        boundary_face_function_numbers[0] = 0
        boundary_face_function_numbers[1] = 1
        boundary_face_physical_tags = np.array([0, 1], dtype=int)
        boundary_face_face_indices = np.array([0, n_faces - 1], dtype=int)

        face_cells = np.empty((n_faces, 2), dtype=int)
        # face_cell_face_index = (n_faces + 1)*np.ones((n_faces, 2), dtype=int)
        face_cells[1: n_faces - 1, 0] = list(range(0, n_inner_cells - 1))
        face_cells[1: n_faces - 1, 1] = list(range(1, n_inner_cells))
        # face_cell_face_index[1:n_faces-1, 0] = 1
        # face_cell_face_index[1:n_faces-1, 1] = 0
        face_cells[0, 0] = n_inner_cells
        # face_cell_face_index[0, 0] = 0
        face_cells[0, 1] = 0
        # face_cell_face_index[0, 1] = 0
        face_cells[-1, 0] = n_inner_cells - 1
        # face_cell_face_index[-1, 0] = 1
        face_cells[-1, 1] = n_inner_cells + 1
        # face_cell_face_index[-1, 1] = 0
        face_centers = 0.5 * (
            cell_centers[face_cells[:, 0]] + cell_centers[face_cells[:, 1]]
        )

        face_subvolumes = np.empty((n_faces, 2), dtype=float)
        face_subvolumes[:, 0] = dx / 2
        face_subvolumes[:, 1] = dx / 2

        boundary_conditions_sorted_physical_tags = np.array([0, 1], dtype=int)
        boundary_conditions_sorted_names = np.array(["left", "right"])

        lsq_gradQ = np.zeros((n_cells, dimension, n_cells), dtype=float)
        deltaQ = np.zeros((n_cells, n_faces_per_cell, n_cells), dtype=float)
        cell_centers = cell_centers[:, :dimension]

        polynomial_degree = 1
        n_neighbors = n_faces_per_cell * polynomial_degree
        dim = 1
        lsq_gradQ, lsq_neighbors, lsq_monomial_multi_index = (
            least_squares_reconstruction_local(
                n_cells, dim, cell_neighbors, cell_centers, lsq_degree
            )
        )
        lsq_scale_factors = scale_lsq_derivative(lsq_monomial_multi_index)

        n_face_neighbors = 2
        face_neighbors = (n_cells + 1) * \
            np.ones((n_faces, n_face_neighbors), dtype=int)

        for i_f, neighbors in enumerate(face_cells):
            face_neighbors[i_f] = neighbors

        z_ordering = np.array([-1], dtype=float)

        # return cls(dimension, 'line', n_cells, n_cells + 1, 2, n_faces_per_element, vertex_coordinates, element_vertices, element_face_areas, element_centers, element_volume, element_inradius, element_face_normals, element_n_neighbors, element_neighbors, element_neighbors_face_index, boundary_face_vertices, boundary_face_corresponding_element, boundary_face_element_face_index, boundary_face_tag, boundary_tag_names)
        return cls(
            dimension,
            "line",
            n_cells,
            n_inner_cells,
            n_faces,
            n_vertices,
            n_boundary_faces,
            n_faces_per_cell,
            vertex_coordinates.T,
            cell_vertices.T,
            cell_faces.T,
            cell_volumes,
            cell_centers.T,
            cell_inradius,
            cell_neighbors,
            boundary_face_cells.T,
            boundary_face_ghosts.T,
            boundary_face_function_numbers,
            boundary_face_physical_tags,
            boundary_face_face_indices.T,
            face_cells.T,
            face_normals.T,
            face_volumes,
            face_centers,
            face_subvolumes,
            face_neighbors,
            boundary_conditions_sorted_physical_tags,
            boundary_conditions_sorted_names,
            lsq_gradQ,
            lsq_neighbors,
            lsq_monomial_multi_index,
            lsq_scale_factors,
            z_ordering,
        )

    def _compute_ascending_order_structured_axis(self):
        cell_centers = self.cell_centers.T
        dimension = self.dimension
        if dimension == 1:
            order = np.lexsort((cell_centers[:, 0],))
        elif dimension == 2:
            order = np.lexsort(
                (
                    cell_centers[:, 0],
                    cell_centers[:, 1],
                )
            )
        elif dimension == 3:
            order = np.lexsort(
                (
                    cell_centers[:, 0],
                    cell_centers[:, 1],
                    cell_centers[:, 2],
                )
            )
        else:
            assert False

        # find number of cells in z (Nz) from first column
        Nz = 1
        while cell_centers[order[Nz], 2] > cell_centers[order[Nz - 1], 2]:
            Nz += 1
        Nz += 1

        # partition order into [(Nx x Ny) , (Nz)]
        N = int(cell_centers.shape[0] / Nz)
        assert cell_centers.shape[0] == N * Nz
        order = order.reshape((N, Nz))

        coords_z = order[0]
        for o in order:
            assert o == coords_z

        return order

    @classmethod
    def from_gmsh(cls, filepath, allow_z_integration=False, lsq_degree=1):
        if not _HAVE_PETSC:
            raise RuntimeError(
                "Mesh.from_gmsh() requires petsc4py, which is not available."
            )
        dm = PETSc.DMPlex().createFromFile(filepath, comm=PETSc.COMM_WORLD)
        boundary_dict = get_physical_boundary_labels(filepath)
        (cStart, cEnd) = dm.getHeightStratum(0)
        (vStart, vEnd) = dm.getDepthStratum(0)
        # (eStart, eEnd) = dm.getDepthStratum(1)
        (eStart, eEnd) = dm.getHeightStratum(1)
        gdm = dm.clone()
        gdm.constructGhostCells()
        (cgStart, cgEnd) = gdm.getHeightStratum(0)
        (vgStart, vgEnd) = gdm.getDepthStratum(0)
        # (egStart, egEnd) = gdm.getDepthStratum(1)
        (egStart, egEnd) = gdm.getHeightStratum(1)

        dim = dm.getDimension()
        n_faces_per_cell = len(gdm.getCone(cgStart))
        n_vertices_per_cell = n_faces_per_cell
        if dim > 2:
            transitive_closure_points, transitive_closure_orientation = (
                dm.getTransitiveClosure(cStart, useCone=True)
            )
            n_vertices_per_cell = transitive_closure_points[
                np.logical_and(
                    transitive_closure_points >= vStart,
                    transitive_closure_points < vEnd,
                )
            ].shape[0]
        n_cells = cgEnd - cgStart
        n_inner_cells = cEnd - cStart
        n_faces = egEnd - egStart
        n_vertices = vEnd - vStart
        cell_vertices = np.zeros(
            (n_inner_cells, n_vertices_per_cell), dtype=int)
        cell_faces = np.zeros((n_inner_cells, n_faces_per_cell), dtype=int)
        cell_centers = np.zeros((n_cells, 3), dtype=float)
        # I create cell_volumes of size n_cells because then I can avoid an if clause in the numerical flux computation. The values will be delted after using apply_boundary_conditions anyways
        cell_volumes = np.ones((n_cells), dtype=float)
        cell_inradius = compute_cell_inradius(dm)
        for i_c, c in enumerate(range(cStart, cEnd)):
            cell_volume, cell_center, cell_normal = dm.computeCellGeometryFVM(
                c)
            transitive_closure_points, transitive_closure_orientation = (
                dm.getTransitiveClosure(c, useCone=True)
            )
            _cell_vertices = transitive_closure_points[
                np.logical_and(
                    transitive_closure_points >= vStart,
                    transitive_closure_points < vEnd,
                )
            ]
            # _cell_vertices_orientation = transitive_closure_orientation[np.logical_and(transitive_closure_points >= vStart, transitive_closure_points <vEnd)]
            assert _cell_vertices.shape[0] == cell_vertices.shape[1]
            # assert (_cell_vertices_orientation == 0).all()
            cell_vertices[i_c, :] = _cell_vertices - vStart
            cell_centers[i_c, :dim] = cell_center[:dim]
            cell_volumes[i_c] = cell_volume

        vertex_coordinates = np.array(gdm.getCoordinates()).reshape((-1, dim))
        boundary_face_cells = {k: [] for k in boundary_dict.values()}
        boundary_face_ghosts = {k: [] for k in boundary_dict.values()}
        boundary_face_face_indices = {k: [] for k in boundary_dict.values()}
        boundary_face_physical_tags = {k: [] for k in boundary_dict.values()}
        face_cells = []
        # face_cell_face_index = np.zeros((n_faces, 2), dtype=int)
        face_normals = []
        face_volumes = []
        face_centers = []
        face_subvolumes = []
        allowed_keys = []
        vertex_coordinates = np.array(dm.getCoordinates()).reshape((-1, dim))

        def get_face_vertices(dim, gdm, vgStart, e):
            if dim == 2:
                return gdm.getCone(e) - vgStart
            elif dim == 3:
                face_vertices = set()
                face_edges = gdm.getCone(e)
                for edge in face_edges:
                    face_vertices.update(gdm.getCone(edge))
                return np.array(list(face_vertices), dtype=int) - vgStart
            else:
                assert False

        for e in range(egStart, egEnd):
            label = gdm.getLabelValue("Face Sets", e)
            face_volume, face_center, face_normal = gdm.computeCellGeometryFVM(
                e)
            face_vertices = get_face_vertices(dim, gdm, vgStart, e)
            face_vertices_coords = vertex_coordinates[face_vertices]
            _face_cells = gdm.getSupport(e)

            _, _cell_center, _ = gdm.computeCellGeometryFVM(_face_cells[0])
            _face_subvolume = np.zeros(2, dtype=float)
            _face_subvolume[0] = compute_subvolume(
                face_vertices_coords, _cell_center, dim
            )
            if _face_cells[1] < n_inner_cells:
                _, _cell_center, _ = gdm.computeCellGeometryFVM(_face_cells[1])
                _face_subvolume[1] = compute_subvolume(
                    face_vertices_coords, _cell_center, dim
                )

            # 2 cells support an face. Ghost cell is the one with the higher number
            if label > -1:
                allowed_keys.append(label)
                boundary_cell = gdm.getSupport(e).min() - cStart
                boundary_ghost = gdm.getSupport(e).max() - cStart
                boundary_face_cells[label].append(boundary_cell)
                boundary_face_ghosts[label].append(boundary_ghost)
                boundary_face_face_indices[label].append(e - egStart)
                boundary_face_physical_tags[label].append(label)
                # boundary_face_vertices[label].append(gdm.getCone(e)-vStart)
                # for periodic boudnary conditions, I need the ghost cell to have a cell_center. I copy the one from the related inner cell.
                _face_cell = gdm.getSupport(e).min()
                _face_ghost = gdm.getSupport(e).max()
                cell_centers[_face_ghost, :dim] = (
                    cell_centers[_face_cell, :dim]
                    + 2
                    * ((face_center - cell_centers[_face_cell, :dim]) @ face_normal)
                    * face_normal
                )
                # subvolumes of the ghost cell are computes wrongly. In this case, copy the value from the inner cell.
                if _face_cells[0] > _face_cells[1]:
                    _face_subvolume[0] = _face_subvolume[1]
                else:
                    _face_subvolume[1] = _face_subvolume[0]

            face_centers.append(face_center)
            _face_cells = gdm.getSupport(e) - cStart
            face_volumes.append(face_volume)
            face_normals.append(face_normal)
            face_cells.append(_face_cells)
            face_subvolumes.append(_face_subvolume)

        # I only want to iterate over the inner cells, but in the ghosted mesh
        for i_c, c in enumerate(range(cgStart, cgStart + n_inner_cells)):
            faces = gdm.getCone(c) - egStart
            cell_faces[i_c, :] = faces

        # least squares liner reconstruction
        # Consider 2d, three points, scalar field, then the reconstruction is for a quad mesh
        # DU = [u_{i+1, j} - u_{i,j}, u_{i-1, j} - u_{i,j}, u_{i, j+1} - u_{i,j}, u_{i, j-1} - u_{i,j}]  \in \mathbb{R}^{4}
        # DX = [[x_{i+1, j} - x_{i,j}, x_{i-1, j} - x_{i,j}, x_{i, j+1} - x_{i,j}, x_{i, j-1} - x_{i,j}],
        #       [y_{i+1, j} - y_{i,j}, y_{i-1, j} - y_{i,j}, y_{i, j+1} - y_{i,j}, y_{i, j-1} - y_{i,j}]]  \in \mathbb{R}^{4x2}
        # S = [S_x, S_y] \mathbb{R}^{2}
        # solve DU = DX \dot S via normal equation
        # A = (DX.T \dot DX)^{-1} DX.T \mathbb{R}^{2x4}
        # so the solution is given by S = A \dot DU
        # the vectorized version is more complicated. I have the vectorization (...) in the first dimension of size n_cells
        # However, I do not want to apply the matrix vector product on DU (such that I can get the reconstruction with a single matrix vector product in the solver), but rather on a scalar field q \in \mathbb{R}^{n_cells}. This requires the need of a discretization matrix D \in \mathbb{R}^{n_cells x 4 x n_cells}, such that Dq = DU \in \mathbb{n_cells x 4}, where the first dimension is the dimension of vectorization

        # cell_neighbors = np.zeros(1)
        lsq_gradQ = np.zeros(1)
        # NON_VECTORIZED CASE
        polynomial_degree = 1
        n_neighbors = n_faces_per_cell * polynomial_degree
        cell_neighbors = (n_cells + 1) * \
            np.ones((n_cells, n_neighbors), dtype=int)

        for i_c, c in enumerate(range(cgStart, cgEnd)):
            # GET NEIGHBORHOOD
            neighbors = _get_neighberhood(gdm, c, cStart=cgStart)
            assert not (i_c == neighbors).any()
            _n_neighbors = neighbors.shape[0]
            if _n_neighbors == 1:
                neighbors_of_neighbor = _get_neighberhood(
                    gdm, neighbors[0] + cgStart, cStart=cgStart
                )
                assert len(neighbors_of_neighbor) == n_faces_per_cell
                neighbors = np.setdiff1d(
                    np.union1d(neighbors_of_neighbor, neighbors), [c]
                )
            cell_neighbors[i_c, :] = neighbors

        lsq_gradQ, lsq_neighbors, lsq_monomial_multi_index = (
            least_squares_reconstruction_local(
                n_cells, dim, cell_neighbors, cell_centers[:, :dim], lsq_degree
            )
        )
        lsq_scale_factors = scale_lsq_derivative(lsq_monomial_multi_index)

        n_face_neighbors = (2 * (n_faces_per_cell + 1) - 2) * polynomial_degree
        face_neighbors = (n_cells + 1) * \
            np.ones((n_faces, n_face_neighbors), dtype=int)

        for i_f, f in enumerate(range(egStart, egEnd)):
            # GET NEIGHBORHOOD
            neighbors = _fill_neighborhood(
                gdm, gdm.getSupport(f), n_face_neighbors, cgStart
            )
            face_neighbors[i_f, :] = neighbors

        face_volumes = np.array(face_volumes, dtype=float)
        _face_centers = np.array(face_centers, dtype=float)
        face_centers = np.zeros((n_faces, 3), dtype=float)
        face_centers[:, :dim] = _face_centers[:, :dim]
        face_normals = np.array(face_normals, dtype=float)
        face_subvolumes = np.array(face_subvolumes, dtype=float)

        face_cells = np.array(face_cells, dtype=int)
        # face_cell_face_index = np.array(face_cell_face_index, dtype=int)
        boundary_face_function_numbers = _boundary_dict_indices(
            boundary_face_cells)

        # get rid of empty keys in the boundary_dict (e.g. no surface values in 2d)
        boundary_dict_inverted = {v: k for k, v in boundary_dict.items()}
        boundary_dict_reduced = {
            k: boundary_dict_inverted[k] for k in allowed_keys}

        # sort the dict by the values
        sorted_keys = np.array(list(boundary_dict_reduced.keys()), dtype=int)
        sorted_keys.sort()
        boundary_dict = {k: boundary_dict_reduced[k] for k in sorted_keys}
        boundary_face_cells = {k: boundary_face_cells[k] for k in sorted_keys}
        boundary_face_ghosts = {
            k: boundary_face_ghosts[k] for k in sorted_keys}
        boundary_face_face_indices = {
            k: boundary_face_face_indices[k] for k in sorted_keys
        }

        boundary_conditions_sorted_physical_tags = np.array(
            list(boundary_dict.keys()), dtype=int
        )
        boundary_conditions_sorted_names = np.array(
            list(boundary_dict.values()), dtype="str"
        )
        boundary_face_cells = np.array(
            _boundary_dict_to_list(boundary_face_cells), dtype=int
        )
        boundary_face_ghosts = np.array(
            _boundary_dict_to_list(boundary_face_ghosts), dtype=int
        )
        boundary_face_physical_tags = np.array(
            _boundary_dict_to_list(boundary_face_physical_tags), dtype=int
        )
        boundary_face_face_indices = np.array(
            _boundary_dict_to_list(boundary_face_face_indices), dtype=int
        )
        n_boundary_faces = boundary_face_cells.shape[0]

        mesh_type = get_mesh_type_from_dm(n_faces_per_cell, dim)

        z_ordering = np.array([-1], dtype=float)

        obj = cls(
            dim,
            mesh_type,
            n_cells,
            n_inner_cells,
            n_faces,
            n_vertices,
            n_boundary_faces,
            n_faces_per_cell,
            vertex_coordinates.T,
            cell_vertices.T,
            cell_faces.T,
            cell_volumes,
            cell_centers.T,
            cell_inradius,
            cell_neighbors,
            boundary_face_cells.T,
            boundary_face_ghosts.T,
            boundary_face_function_numbers,
            boundary_face_physical_tags,
            boundary_face_face_indices.T,
            face_cells.T,
            face_normals.T,
            face_volumes,
            face_centers,
            face_subvolumes,
            face_neighbors,
            boundary_conditions_sorted_physical_tags,
            boundary_conditions_sorted_names,
            lsq_gradQ,
            lsq_neighbors,
            lsq_monomial_multi_index,
            lsq_scale_factors,
            z_ordering,
        )

        if allow_z_integration:
            obj._compute_ascending_order_structured_axis()

        return obj

    @classmethod
    def extrude_mesh(cls, msh, n_layers=10):
        Z = np.linspace(0, 1, n_layers + 1)
        dimension = msh.dimension + 1
        mesh_type = get_extruded_mesh_type(msh.type)
        n_cells = msh.n_cells * n_layers
        n_inner_cells = msh.n_inner_cells * n_layers
        n_vertices = msh.n_cells * (n_layers + 1)
        n_boundary_faces = msh.n_boundary_faces * n_layers + 2 * msh.n_cells
        n_faces_per_cell = mesh_util._get_faces_per_element(mesh_type)
        n_faces = n_inner_cells * n_faces_per_cell
        vertex_coordinates = extrude.extrude_points(
            msh.vertex_coordinates.T, Z).T
        cell_vertices = extrude.extrude_element_vertices(
            msh.cell_vertices.T, msh.n_vertices, n_layers
        ).T

        cell_centers = np.empty((n_cells, 3), dtype=float)
        cell_volumes = np.empty((n_cells), dtype=float)
        cell_inradius = np.empty((n_cells), dtype=float)
        cell_face_areas = np.empty((n_cells, n_faces_per_cell), dtype=float)
        cell_face_normals = np.zeros(
            (n_cells, n_faces_per_cell, 3), dtype=float)
        cell_n_neighbors = np.empty((n_cells), dtype=int)
        cell_neighbors = np.empty((n_cells, n_faces_per_cell), dtype=int)
        cell_neighbors_face_index = np.empty(
            (n_cells, n_faces_per_cell), dtype=int)
        for i_elem, elem in enumerate(cell_vertices.T):
            cell_centers[i_elem, :dimension] = mesh_util.center(
                vertex_coordinates.T, elem
            )

        # truncate normals and positions from 3d to dimendion-d
        vertex_coordinates = vertex_coordinates.T[:, :dimension].T
        cell_centers = cell_centers[:, :]
        # cell_face_normals = cell_face_normals[:, :, :dimension]

        # TODO
        # empty fields
        cell_faces = np.empty((n_inner_cells, n_faces_per_cell), dtype=int)
        boundary_face_cells = np.array([0, n_inner_cells - 1], dtype=int)
        boundary_face_ghosts = np.array(
            [n_inner_cells, n_inner_cells + 1], dtype=int)
        boundary_face_function_numbers = np.empty(
            (n_boundary_faces), dtype=int)
        boundary_face_physical_tags = np.array([0, 1], dtype=int)
        boundary_face_face_indices = np.array([0, n_faces - 1], dtype=int)
        face_cells = np.empty((n_faces, 2), dtype=int)
        face_subvolumes = np.empty((n_faces, 2), dtype=float)
        boundary_conditions_sorted_physical_tags = np.array([0, 1], dtype=int)
        boundary_conditions_sorted_names = np.array(["left", "right"])
        face_normals = np.zeros((n_faces, 3), dtype=float)
        face_volumes = np.zeros((n_faces), dtype=float)
        face_centers = np.zeros((n_faces, 3), dtype=float)

        # hard coded guess
        n_face_neighbors = 0
        face_neighbors = (n_cells + 1) * \
            np.ones((n_faces, n_face_neighbors), dtype=int)
        lsq_gradQ = np.zeros((n_cells, dimension, 0), dtype=float)
        lsq_neighbors = np.zeros(1)
        lsq_monomial_multi_index = np.zeros(1)
        lsq_scale_factors = np.zeros(1)
        z_ordering = np.array([-1], dtype=float)

        return cls(
            msh.dimension + 1,
            mesh_type,
            n_cells,
            n_inner_cells,
            n_faces,
            n_vertices,
            n_boundary_faces,
            n_faces_per_cell,
            vertex_coordinates,
            cell_vertices,
            cell_faces.T,
            cell_volumes,
            cell_centers.T,
            cell_inradius,
            cell_neighbors,
            boundary_face_cells.T,
            boundary_face_ghosts.T,
            boundary_face_function_numbers,
            boundary_face_physical_tags,
            boundary_face_face_indices.T,
            face_cells.T,
            face_normals.T,
            face_volumes,
            face_centers,
            face_subvolumes,
            face_neighbors,
            boundary_conditions_sorted_physical_tags,
            boundary_conditions_sorted_names,
            lsq_gradQ,
            lsq_neighbors,
            lsq_monomial_multi_index,
            lsq_scale_factors,
            z_ordering,
        )

    def write_to_hdf5(self, filepath: str):
        if not _HAVE_H5PY:
            raise RuntimeError(
                "Mesh.write_to_hdf5() requires h5py, which is not available."
            )
        main_dir = os.getenv("ZOOMY_DIR")
        with h5py.File(os.path.join(main_dir, filepath), "w") as f:
            mesh = f.create_group("mesh")
            mesh.create_dataset("dimension", data=self.dimension)
            mesh.create_dataset("type", data=self.type)
            mesh.create_dataset("n_cells", data=self.n_cells)
            mesh.create_dataset("n_inner_cells", data=self.n_inner_cells)
            mesh.create_dataset("n_faces", data=self.n_faces)
            mesh.create_dataset("n_vertices", data=self.n_vertices)
            mesh.create_dataset("n_boundary_faces", data=self.n_boundary_faces)
            mesh.create_dataset("n_faces_per_cell", data=self.n_faces_per_cell)
            mesh.create_dataset("vertex_coordinates",
                                data=self.vertex_coordinates)
            mesh.create_dataset("cell_vertices", data=self.cell_vertices)
            mesh.create_dataset("cell_faces", data=self.cell_faces)
            mesh.create_dataset("cell_volumes", data=self.cell_volumes)
            mesh.create_dataset("cell_centers", data=self.cell_centers)
            mesh.create_dataset("cell_inradius", data=self.cell_inradius)
            mesh.create_dataset("cell_neighbors", data=self.cell_neighbors)
            mesh.create_dataset("boundary_face_cells",
                                data=self.boundary_face_cells)
            mesh.create_dataset("boundary_face_ghosts",
                                data=self.boundary_face_ghosts)
            mesh.create_dataset(
                "boundary_face_function_numbers",
                data=self.boundary_face_function_numbers,
            )
            mesh.create_dataset(
                "boundary_face_physical_tags", data=self.boundary_face_physical_tags
            )
            mesh.create_dataset(
                "boundary_face_face_indices", data=self.boundary_face_face_indices
            )
            mesh.create_dataset("face_cells", data=self.face_cells)
            # mesh.create_dataset("face_cell_face_index", data=self.face_cell_face_index)
            mesh.create_dataset("face_normals", data=self.face_normals)
            mesh.create_dataset("face_volumes", data=self.face_volumes)
            mesh.create_dataset("face_centers", data=self.face_centers)
            mesh.create_dataset("face_subvolumes", data=self.face_subvolumes)
            mesh.create_dataset("face_neighbors", data=self.face_neighbors)
            mesh.create_dataset(
                "boundary_conditions_sorted_physical_tags",
                data=np.array(self.boundary_conditions_sorted_physical_tags),
            )
            mesh.create_dataset(
                "boundary_conditions_sorted_names",
                data=np.array(
                    self.boundary_conditions_sorted_names, dtype="S"),
            )
            mesh.create_dataset("lsq_gradQ", data=np.array(self.lsq_gradQ))
            mesh.create_dataset(
                "lsq_neighbors", data=np.array(self.lsq_neighbors))
            mesh.create_dataset(
                "lsq_monomial_multi_index", data=(self.lsq_monomial_multi_index)
            )
            mesh.create_dataset("lsq_scale_factors",
                                data=(self.lsq_scale_factors))
            mesh.create_dataset("z_ordering", data=np.array(self.z_ordering))

    @classmethod
    def from_hdf5(cls, filepath: str):
        if not _HAVE_H5PY:
            raise RuntimeError(
                "Mesh.from_hdf5() requires h5py, which is not available."
            )
        with h5py.File(filepath, "r") as file:
            file = file
            mesh = cls(
                file["mesh"]["dimension"][()],
                (file["mesh"]["type"][()]).decode("utf-8"),
                file["mesh"]["n_cells"][()],
                file["mesh"]["n_inner_cells"][()],
                file["mesh"]["n_faces"][()],
                file["mesh"]["n_vertices"][()],
                file["mesh"]["n_boundary_faces"][()],
                file["mesh"]["n_faces_per_cell"][()],
                file["mesh"]["vertex_coordinates"][()],
                file["mesh"]["cell_vertices"][()],
                file["mesh"]["cell_faces"][()],
                file["mesh"]["cell_volumes"][()],
                file["mesh"]["cell_centers"][()],
                file["mesh"]["cell_inradius"][()],
                file["mesh"]["cell_neighbors"][()],
                file["mesh"]["boundary_face_cells"][()],
                file["mesh"]["boundary_face_ghosts"][()],
                file["mesh"]["boundary_face_function_numbers"][()],
                file["mesh"]["boundary_face_physical_tags"][()],
                file["mesh"]["boundary_face_face_indices"][()],
                file["mesh"]["face_cells"][()],
                # file["mesh"]["face_cell_face_index"][()],
                file["mesh"]["face_normals"][()],
                file["mesh"]["face_volumes"][()],
                file["mesh"]["face_centers"][()],
                file["mesh"]["face_subvolumes"][()],
                file["mesh"]["face_neighbors"][()],
                file["mesh"]["boundary_conditions_sorted_physical_tags"][()],
                np.array(
                    file["mesh"]["boundary_conditions_sorted_names"][()
                                                                     ], dtype="str"
                ),
                file["mesh"]["lsq_gradQ"][()],
                file["mesh"]["lsq_neighbors"][()],
                file["mesh"]["lsq_monomial_multi_index"][()],
                file["mesh"]["lsq_scale_factors"][()],
                file["mesh"]["z_ordering"][()],
            )
        return mesh

    def write_to_vtk(
        self,
        filepath: str,
        fields: Union[FArray, None] = None,
        field_names: Union[list[str], None] = None,
        point_data: dict = {},
    ):
        if not _HAVE_MESHIO:
            raise RuntimeError(
                "_write_to_vtk_from_vertices_edges requires meshio, which is not available."
            )
        d_fields = {}
        vertex_coords_3d = np.zeros((3, self.vertex_coordinates.shape[1]))
        vertex_coords_3d[: self.vertex_coordinates.shape[0], :] = (
            self.vertex_coordinates
        )
        if fields is not None:
            if field_names is None:
                field_names = [str(i) for i in range(fields.shape[0])]
            for i_fields, field in enumerate(fields):
                d_fields[field_names[i_fields]] = [field]
        # the brackets around the second argument (cells) indicate that I have only mesh type of mesh element of type self.type, with corresponding vertices.
        meshout = meshio.Mesh(
            vertex_coords_3d.T,
            [
                (
                    mesh_util.convert_mesh_type_to_meshio_mesh_type(self.type),
                    self.cell_vertices.T,
                )
            ],
            cell_data=d_fields,
            point_data=point_data,
        )
        path, _ = os.path.split(filepath)
        filepath, file_ext = os.path.splitext(filepath)
        if not os.path.exists(path) and path != "":
            os.mkdir(path)
        meshout.write(filepath + ".vtk", binary=False)


if __name__ == "__main__":
    path = "/home/ingo/Git/sms/meshes/quad_2d/mesh_coarse.msh"
    path2 = "/home/ingo/Git/sms/meshes/quad_2d/mesh_fine.msh"
    path3 = "/home/ingo/Git/sms/meshes/quad_2d/mesh_finest.msh"
    path4 = "/home/ingo/Git/sms/meshes/triangle_2d/mesh_coarse.msh"
    labels = get_physical_boundary_labels(path)

    # dm, boundary_dict, ghost_cells_dict = load_gmsh(path)

    mesh = Mesh.from_gmsh(path)
    assert mesh.cell_faces.max() == mesh.n_faces - 1
    assert mesh.cell_faces.min() == 0
    assert mesh.face_cells.max() == mesh.n_cells - 1
    assert mesh.face_cells.min() == 0
    assert mesh.cell_vertices.max() == mesh.n_vertices - 1
    assert mesh.cell_vertices.min() == 0

    mesh.write_to_hdf5("./test.h5")
    mesh = Mesh.from_hdf5("./test.h5")
    # mesh.write_to_vtk('./test.vtk')
    mesh.write_to_vtk(
        "./test.vtk",
        fields=np.ones((2, mesh.n_inner_cells), dtype=float),
        field_names=["A", "B"],
    )



import numpy as np
import numpy.polynomial.legendre as L
import numpy.polynomial.chebyshev as C
from scipy.optimize import least_squares as lsq
import sympy
from sympy import Matrix, sqrt
from sympy.abc import x
from attr import define, field

from sympy import integrate, diff
from sympy import legendre
from sympy import lambdify

from zoomy_core.misc.misc import Zstruct
from zoomy_core.model.basemodel import (
    register_sympy_attribute,
    eigenvalue_dict_to_matrix,
)
from zoomy_core.model.basemodel import Model
import zoomy_core.model.initial_conditions as IC


@define(kw_only=True, slots=True, frozen=True)
class Poisson(Model):
    dimension: int = 1
    variables: Zstruct = field(init=False, default=1)
    aux_variables: Zstruct = field(factory = lambda: ['ddTdxx', 'ddTdyy', 'ddTdzz'])
    
            
    def residual(self):
        R = Matrix([0 for i in range(self.n_variables)])
        T = self.variables[0]
        ddTdxx = self.aux_variables.ddTdxx
        param = self.parameters

        R[0] = - ddTdxx + 2
        return R
        

    
    



from typing import Callable
import numpy as np
from attrs import define, field

from zoomy_core.misc.custom_types import FArray


@define(kw_only=False, slots=True, frozen=True)
class NumpyRuntimeModel:
    """Runtime model generated from a symbolic Model."""

    # --- User-provided ---
    model: object = field()

    # --- Automatically derived ---
    name: str = field(init=False)
    dimension: int = field(init=False)
    n_variables: int = field(init=False)
    n_aux_variables: int = field(init=False)
    n_parameters: int = field(init=False)
    parameters: FArray = field(init=False)
    flux: Callable = field(init=False)
    dflux: Callable = field(init=False)
    source: Callable = field(init=False)
    source_jacobian_wrt_variables: Callable = field(init=False)
    source_jacobian_wrt_aux_variables: Callable = field(init=False)
    nonconservative_matrix: Callable = field(init=False)
    quasilinear_matrix: Callable = field(init=False)
    eigenvalues: Callable = field(init=False)
    residual: Callable = field(init=False)
    project_2d_to_3d: Callable = field(init=False)
    project_3d_to_2d: Callable = field(init=False)
    boundary_conditions: Callable = field(default=None, init=False)

    left_eigenvectors: Callable = field(default=None, init=False)
    right_eigenvectors: Callable = field(default=None, init=False)

    # --- Constants ---
    module = {
        "ones_like": np.ones_like,
        "zeros_like": np.zeros_like,
        "array": np.array,
        "squeeze": np.squeeze,
    }
    printer = "numpy"
    

    # ------------------------------------------------------------------

    def __attrs_post_init__(self):
        model = self.model

        
        flux=model._flux.lambdify(modules=[self.module, self.printer])
        dflux=model._dflux.lambdify(modules=[self.module, self.printer])
        nonconservative_matrix=model._nonconservative_matrix.lambdify(
            modules=[self.module, self.printer]
        )
        quasilinear_matrix=model._quasilinear_matrix.lambdify(
            modules=[self.module, self.printer]
        )
        eigenvalues=model._eigenvalues.lambdify(
            modules=[self.module, self.printer]
        )
        left_eigenvectors=model._left_eigenvectors.lambdify(
            modules=[self.module, self.printer]
        )
        right_eigenvectors=model._right_eigenvectors.lambdify(
            modules=[self.module, self.printer]
        )
        source=model._source.lambdify(modules=[self.module, self.printer])
        source_jacobian_wrt_variables=model._source_jacobian_wrt_variables.lambdify(
            modules=[self.module, self.printer]
        )
        source_jacobian_wrt_aux_variables=model._source_jacobian_wrt_aux_variables.lambdify(
            modules=[self.module, self.printer]
        )
        residual=model._residual.lambdify(modules=[self.module, self.printer])
        project_2d_to_3d=model._project_2d_to_3d.lambdify(
            modules=[self.module, self.printer]
        )
        project_3d_to_2d=model._project_3d_to_2d.lambdify(
            modules=[self.module, self.printer]
        )
        bcs = model._boundary_conditions.lambdify(modules=[self.module, self.printer])


        # --- Assign frozen fields -------------------------------------
        object.__setattr__(self, "name", model.name)
        object.__setattr__(self, "dimension", model.dimension)
        object.__setattr__(self, "n_variables", model.n_variables)
        object.__setattr__(self, "n_aux_variables", model.n_aux_variables)
        object.__setattr__(self, "n_parameters", model.n_parameters)
        object.__setattr__(self, "parameters", model.parameter_values)

        object.__setattr__(self, "flux", flux)
        object.__setattr__(self, "dflux", dflux)
        object.__setattr__(self, "source", source)
        object.__setattr__(
            self, "source_jacobian_wrt_variables", source_jacobian_wrt_variables
        )
        object.__setattr__(
            self,
            "source_jacobian_wrt_aux_variables",
            source_jacobian_wrt_aux_variables,
        )
        object.__setattr__(self, "nonconservative_matrix", nonconservative_matrix)
        object.__setattr__(self, "quasilinear_matrix", quasilinear_matrix)
        object.__setattr__(self, "eigenvalues", eigenvalues)
        object.__setattr__(self, "left_eigenvectors", left_eigenvectors)
        object.__setattr__(self, "right_eigenvectors", right_eigenvectors)
        object.__setattr__(self, "residual", residual)
        object.__setattr__(self, "project_2d_to_3d", project_2d_to_3d)
        object.__setattr__(self, "project_3d_to_2d", project_3d_to_2d)
        object.__setattr__(self, "boundary_conditions", bcs)



from copy import deepcopy

import numpy as np
import sympy
from sympy import Symbol, bspline_basis_set, diff, integrate, lambdify, legendre
from sympy.abc import z
from sympy.functions.special.polynomials import chebyshevu


class Basisfunction:
    name = "Basisfunction"
    
    def bounds(self):
        return [0, 1]

    def basis_definition(self):
        z = Symbol("z")
        b = lambda k, z: z**k
        return [b(k, z) for k in range(self.level + 1)]
    
    def weight(self, z):
        return 1
    
    def weight_eval(self, z):
        z = Symbol("z")
        f = sympy.lambdify(z, self.weight(z))
        return f(z)

    def __init__(self, level=0, **kwargs):
        self.level = level
        self.basis = self.basis_definition(**kwargs)

    def get(self, k):
        return self.basis[k]

    def eval(self, k, _z):
        return self.get(k).subs(z, _z)
    
    def eval_psi(self, k, _z):
        z = sympy.Symbol('z')
        psi = sympy.integrate(self.get(k), (z, self.bounds()[0], z))
        return psi.subs(z, _z)

    def get_lambda(self, k):
        f = lambdify(z, self.get(k))

        def lam(z):
            if type(z) == int or type(z) == float:
                return f(z)
            elif type(z) == list or type(z) == np.ndarray:
                return np.array([f(xi) for xi in z])
            else:
                assert False

        return lam

    def plot(self, ax):
        X = np.linspace(self.bounds()[0], self.bounds()[1], 1000)
        for i in range(len(self.basis)):
            f = lambdify(z, self.get(i))
            y = np.array([f(xi) for xi in X])
            ax.plot(X, y, label=f"basis {i}")

    def reconstruct_velocity_profile(self, alpha, N=100):
        Z = np.linspace(self.bounds()[0], self.bounds()[1], N)
        u = np.zeros_like(Z)
        for i in range(len(self.basis)):
            b = lambdify(z, self.get(i))
            u[:] += alpha[i] * b(Z)
        return u
    
    def reconstruct_velocity_profile_at(self, alpha, z):
        u = 0
        for i in range(len(self.basis)):
            b = lambdify(z, self.eval(i, z))
            u += alpha[i] * b(z)
        return u

    def reconstruct_alpha(self, velocities, z):
        n_basis = len(self.basis)
        alpha = np.zeros(n_basis)
        for i in range(n_basis):
            b = lambdify(z, self.eval(i, z))
            nom = np.trapz(velocities * b(z) * self.weight(z), z)
            if type(b(z)) == int:
                den = b(z) ** 2
            else:
                den = np.trapz((b(z) * b(z)).reshape(z.shape), z)
            res = nom / den
            alpha[i] = res
        return alpha
    
    def project_onto_basis(self, Y):
        Z = np.linspace(self.bounds()[0], self.bounds()[1], Y.shape[0])
        n_basis = len(self.basis)
        alpha = np.zeros(n_basis)
        z = Symbol("z")
        for i in range(n_basis):
            b = lambdify(z, self.eval(i, Z))
            alpha[i] = np.trapz(Y * b(Z) * self.weight_eval(Z), Z)
        return alpha

    def get_diff_basis(self):
        db = [diff(b, z) for i, b in enumerate(self.basis)]
        return db


class Monomials(Basisfunction):
    name = "Monomials"


class Legendre_shifted(Basisfunction):
    name = "Legendre_shifted"

    def basis_definition(self):
        z = Symbol("z")
        b = lambda k, z: legendre(k, 2 * z - 1) * (-1) ** (k)
        return [b(k, z) for k in range(self.level + 1)]
    
class Chebyshevu(Basisfunction):
    name = "Chebyshevu"
    
    def bounds(self):
        return [-1, 1]
    
    def weight(self, z):
        # do not forget to include the jacobian of the coordinate transformation in the weight
        return sympy.sqrt(1-z**2)

    def basis_definition(self):
        z = Symbol("z")
        b = lambda k, z: sympy.sqrt(2 / sympy.pi) * chebyshevu(k, z)
        return [b(k, z) for k in range(self.level + 1)]
    
class Legendre_DN(Basisfunction):
    name = "Legendre_DN - satifying no-slip and no-stress. This is a non-SWE basis"

    def bounds(self):
        return [-1, 1]

    def basis_definition(self):
        z = Symbol("z")
        def b(k, z):
            alpha = sympy.Rational((2*k+3), (k+2)**2)
            beta = -sympy.Rational((k+1),(k+2))**2
            return (legendre(k, z) ) + alpha * (legendre(k+1, z) ) + beta * (legendre(k+2, z))
        #normalizing makes no sence, as b(k, 0) = 0 by construction
        return [b(k, z) for k in range(self.level + 1)]


class Spline(Basisfunction):
    name = "Spline"

    def basis_definition(self, degree=1, knots=[0, 0, 0.001, 1, 1]):
        z = Symbol("z")
        basis = bspline_basis_set(degree, knots, z)
        return basis


class OrthogonalSplineWithConstant(Basisfunction):
    name = "OrthogonalSplineWithConstant"

    def basis_definition(self, degree=1, knots=[0, 0, 0.5, 1, 1]):
        z = Symbol("z")

        def prod(u, v):
            return integrate(u * v, (z, 0, 1))

        basis = bspline_basis_set(degree, knots, z)
        add_basis = [1]
        # add_basis = [sympy.Piecewise((0, z<0.1), (1, True))]
        basis = add_basis + basis[:-1]
        orth = deepcopy(basis)
        for i in range(1, len(orth)):
            for j in range(0, i):
                orth[i] -= prod(basis[i], orth[j]) / prod(orth[j], orth[j]) * orth[j]

        for i in range(len(orth)):
            orth[i] /= sympy.sqrt(prod(orth[i], orth[i]))

        return orth
